/****************************************************************************
  Project: NITRO-SampleTools - TXLib
  File:    TXLib_ByteData.cpp

  Copyright 2004,2005 Nintendo Co.,Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  Class for storing more than 1-byte of data 

  $Log: TXLib_ByteData.cpp,v $
  Revision 1.7  2005/06/20 06:54:11  mominoki
  Changed copyright. Made revision adjustments. 

  Revision 1.1  2004/04/15		mominoki
  Corrected comment

  Revision 1.0  2004/03/03		mominoki
  Inital registration

  $NoKeywords: $
 ***************************************************************************/
#define TLB_EXPORT_

#include <cstring>
#include <cassert>
#include <TXLib/Base/TXLib_ByteData.h>

namespace std{ using :: memset; }
namespace std{ using :: memcpy; }

namespace TXLib{

/*********************************************************************************************************
	TXLib::CByteData::CByteData

	Constructor
*********************************************************************************************************/
CByteData::CByteData()
:
	m_available(false),
	m_byteSize(0),
	m_byteCount(0),
	m_number(0),
	m_pData( NULL )
{
}

/*********************************************************************************************************
	TXLib::CByteData::CByteData

	Copy Constructor
*********************************************************************************************************/
CByteData::CByteData(
	const CByteData& rByteData	// CByteData Pointer to copy source
) :
	m_available	(rByteData.m_available),
	m_byteSize(rByteData.m_byteSize),
	m_byteCount(rByteData.m_byteCount),
	m_number(rByteData.m_number),
	m_pData( NULL )
{
	if (m_byteSize > 0)
	{
		m_pData	= new u8[ rByteData.m_byteSize ];				// Memory reservation
		std::memcpy( m_pData, rByteData.m_pData, m_byteSize );	// Data Copy
	}
}

/*********************************************************************************************************
	TXLib::CByteData::operator=

	Assignment Operator
*********************************************************************************************************/
CByteData&
CByteData::operator=(
	const CByteData& rByteData		// CByteData Pointer to copy source
)
{
	if (&rByteData != this)
	{
		Clear();							// Clear

		m_available	= rByteData.m_available;
		m_byteSize	= rByteData.m_byteSize;
		m_byteCount = rByteData.m_byteCount;
		m_number	= rByteData.m_number;

		if (m_byteSize > 0)
		{
			m_pData	= new u8[ rByteData.m_byteSize ];			// Memory reservation
			std::memcpy( m_pData, rByteData.m_pData, m_byteSize );	// Data Copy
		}
	}
	return *this;
}

/*********************************************************************************************************
	TXLib::CByteData::~CByteData

	Destructor
*********************************************************************************************************/
CByteData:: ~CByteData()
{
	Clear(); // Clear
}

/*********************************************************************************************************
		private
*********************************************************************************************************/
/*********************************************************************************************************
	TXLib::CByteData::Clear

	Memory Clear
*********************************************************************************************************/
void
CByteData::Clear()
{
	delete[] m_pData;

	m_available = false;
	m_pData = NULL;
	m_byteSize = 0;
	m_byteCount = 0;
	m_number = 0;
}

/*********************************************************************************************************
		public
*********************************************************************************************************/
/*********************************************************************************************************
	TXLib::CByteData::Create

	Create
*********************************************************************************************************/
void
CByteData::Create(
				u8					byteCount,			// Data byte width
				u32					number,				// Data count
				u32					initialData			// Data default value
)
{
	assert( (byteCount == 1) || (byteCount == 2) || (byteCount == 4) );

	Clear();													// Clear

	m_available = true;
	m_byteSize	= byteCount * number;	// Byte size
	m_byteCount = byteCount;			// Byte width
	m_number	= number;				// Number
	m_pData = new u8[ m_byteSize ];		// Memory reservation
	u32 n;

	if( ( byteCount == 1 ) || ( initialData == 0 ))
	{
		std::memset( m_pData, initialData, m_byteSize); // Memory Initialization (Initialize with initialData)
	}
	else if( byteCount == 2 )
	{
		for( n=0; n<number; n++)
		{
			*(u16*)( m_pData + 2 * n ) = initialData;
		}
	}
	else if( byteCount == 4 )
	{
		for( n=0; n<number; n++)
		{
			*(u32*)( m_pData + 4 * n ) = initialData;
		}
	}
}

/*********************************************************************************************************
	TXLib::CByteData::GetValue

	Data ac
*********************************************************************************************************/
inline void
CByteData::GetValue( 
				u32*				pData,				// Data Pointer
				u32					index				// Number
) const
{
	assert(( pData != NULL ) && ( index < m_number ));

	if( m_byteCount == 1)
	{
		*pData = (u32)(*( m_pData + index ));  // Acquisition
	}
	else if( m_byteCount == 2)
	{
		*pData = (u32)( *(u16*)( m_pData + index * 2) );
	}
	else if( m_byteCount == 4)
	{
		*pData = *(u32*)( m_pData + index * 4 );  // Acquisition
	}
	else
	{
		assert(false);
	}
}

/*********************************************************************************************************
	TXLib::CByteData::SetValue

	Set Data
*********************************************************************************************************/
inline void
CByteData::SetValue( 
				u32					data,				// Data
				u32					index				// Number
)
{
	assert( index < m_number );

	if( m_byteCount == 1)
	{
		*( m_pData + index ) = (u8)data;  // Set
	}
	else if( m_byteCount == 2)
	{
		*(u16*)( m_pData + index * 2 ) = (u16)data;  // Set
	}
	else if( m_byteCount == 4)
	{
		*(u32*)( m_pData + index * 4 ) = data;  // Set
	}
	else
	{
		assert(false);
	}
}

} // namespace
