/****************************************************************************
  Project: NITRO-SampleTools - TXLib
  File:    TXLib_ColorPalette.cpp

  Copyright 2004,2005 Nintendo Co.,Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  Class handling color palette compontent of CTexture.
  Do not use when CTexture is Direct.

  $Log: TXLib_ColorPalette.cpp,v $
  Revision 1.7  2005/06/20 06:54:11  mominoki
  Changed copyright. Made revision adjustments. 

  Revision 1.1  2004/04/15		mominoki
  Corrected comment

  Revision 1.0  2004/03/03		mominoki
  Initial registration

  $NoKeywords: $
 ***************************************************************************/
#define TLB_EXPORT_

#include <cstring>
#include <TXLib/Base/TXLib_ColorPalette.h>
#include <TXLib_Define.h>

namespace TXLib{

/*********************************************************************************************************
	TXLib::CColorPalette::

	Constructor
*********************************************************************************************************/
CColorPalette::CColorPalette()
:
m_available(false),
m_redChannel(),
m_greenChannel(),
m_blueChannel()
{
}

/*********************************************************************************************************
	TXLib::CColorPalette::

	Copy constructor
*********************************************************************************************************/
CColorPalette::CColorPalette
(
	const CColorPalette& rColorPalette	// CColorPalette in copy source
) :
	m_available		(rColorPalette.m_available		),
	m_redChannel	(rColorPalette.m_redChannel		),
	m_greenChannel	(rColorPalette.m_greenChannel	),
	m_blueChannel	(rColorPalette.m_blueChannel	)
{
}

/*********************************************************************************************************
	TXLib::CColorPalette::

	Assignment Operator
*********************************************************************************************************/
CColorPalette&
CColorPalette::operator=(
	const CColorPalette& rColorPalette	// CColorPalette in copy source
)
{
	if (&rColorPalette != this)
	{
		m_available	   = rColorPalette.m_available;
		m_redChannel   = rColorPalette.m_redChannel;
		m_greenChannel = rColorPalette.m_greenChannel;
		m_blueChannel  = rColorPalette.m_blueChannel;
	}
	return *this;
}

/*********************************************************************************************************
	TXLib::CColorPalette::

	Destructor

*********************************************************************************************************/
CColorPalette::~CColorPalette()
{ 
	Clear(); // Clear
} 

/*********************************************************************************************************
		private
*********************************************************************************************************/
/*********************************************************************************************************
	TXLib::CColorPalette::

	Memory clear
*********************************************************************************************************/
void
CColorPalette::Clear()
{
	m_available = false;

	m_redChannel  .Clear();
	m_greenChannel.Clear();
	m_blueChannel .Clear();

}

/*********************************************************************************************************
		public
*********************************************************************************************************/
/*********************************************************************************************************
	TXLib::CColorPalette::

	Create

	Create after freeing memory
*********************************************************************************************************/
void
CColorPalette::Create(
	u32					number,				// Data count
	u8					initialRed,			// Default value Red
	u8					initialGreen,		// Default value Green
	u8					initialBlue			// Default value Blue
)
{
	m_available = true;

	m_redChannel  .Create( number, initialRed   );
	m_greenChannel.Create( number, initialGreen );
	m_blueChannel .Create( number, initialBlue  );
}

/*********************************************************************************************************
	TXLib::CColorPalette::

	Get color value
*********************************************************************************************************/
inline void
CColorPalette::GetValue( 
				u8*							pRed,							// Data pointer Red
				u8*							pGreen,							// Data pointer Green
				u8*							pBlue,							// Data pointer Blue
				u32							index							// Number
) const
{
	assert( ( pRed != NULL )  && ( pGreen != NULL ) && ( pBlue  != NULL ) );
	assert( index <= GetNumber() );

	m_redChannel.  GetValue( pRed, index  );
	m_greenChannel.GetValue( pGreen, index  );
	m_blueChannel. GetValue( pBlue, index  );
}

/*********************************************************************************************************
	TXLib::CColorPalette::SetValue

	Set color value
*********************************************************************************************************/
inline void
CColorPalette::SetValue( 
				u8							red,							// Data Red
				u8							green,							// Data Green
				u8							blue,							// Data Blue
				u32							index							// Number
)
{
	assert( index <= GetNumber() );

	m_redChannel.  SetValue( red, index  );
	m_greenChannel.SetValue( green, index  );
	m_blueChannel. SetValue( blue, index  );
}

/*********************************************************************************************************
	TXLib::CColorPalette::CutCommonColor

	Cut common colors

	Cut common colors from color palette, and minimize the number of colors on the palette. 
*********************************************************************************************************/
void
CColorPalette::CutCommonColor()
{
	CData		newRedChannel;		// Data Red
	CData		newGreenChannel;	// Data Green
	CData		newBlueChannel;		// Data Blue

	newRedChannel.  Create( GetNumber(), 0 ); // Initialize
	newGreenChannel.Create( GetNumber(), 0 );
	newBlueChannel. Create( GetNumber(), 0 );

	u32 newNumber=0;			// Number of new colors to make
	u8 red, green, blue;		// Color value

	for( u32 old_n=0; old_n<GetNumber(); old_n++)
	{
		GetValue( &red, &green, &blue, old_n );			// Get original color value 

		// Do the color values exist in new color palette? 
		bool exist = false;								// "Does not exist" initially

		for( u32 new_n=0; new_n<newNumber; new_n++)
		{
			// "Exists"
			if( ( red  == newRedChannel[new_n] ) && 
					( green == newGreenChannel[new_n] ) && ( blue  == newBlueChannel[new_n] ) )
			{
				exist = true;
				break; // Break out of loop
			}
		}

		// If does not exist, enter the color value into new color palette. 
		if( !exist )
		{
			newRedChannel  [ newNumber ] = red;
			newGreenChannel[ newNumber ] = green;
			newBlueChannel [ newNumber]  = blue;
			newNumber++;
		}
	}

	// Replace original palette with new palette
	Create(newNumber, 0, 0, 0 );							// Create

	for( u32 number=0; number<newNumber; number++ )
	{
		SetValue( newRedChannel[ number], newGreenChannel[ number ], newBlueChannel[ number ], number );
	}

	ChangeNumber( newNumber );	// Change count
}

/*********************************************************************************************************
	TXLib::CColorPalette::ChangeNumber

	Change number of colors on palette

	m_number > number: Change the number of colors after setting excessing data to "0."
	m_number < number: Change the number of colors after securing enough memory. Clear the increases to"0."
*********************************************************************************************************/
void
CColorPalette::ChangeNumber( 
				u32							number							// Number of data to change
)
{
	m_redChannel.  ChangeNumber(number);
	m_greenChannel.ChangeNumber(number);
	m_blueChannel. ChangeNumber(number);
}

/*********************************************************************************************************
	TXLib::CColorPalette::FindColor

	Find the specified color from color palette, and return index.

	If it is not found, return -1.
*********************************************************************************************************/
int
CColorPalette::FindColor(
				u8							red,
				u8							green,
				u8							blue
)
{
	for (u32 index=0; index<GetNumber(); index++)
	{
		u8 red2, green2, blue2;

		m_redChannel.  GetValue( &red2,		index );
		m_greenChannel.GetValue( &green2,	index );
		m_blueChannel. GetValue( &blue2,	index );

		if( ( red2 == red ) && ( green2 == green ) && ( blue2 == blue ) )
		{
			return index;
		}
	}

	return  -1;
}

/*********************************************************************************************************
	TXLib::CColorPalette::QuickSortTopDown

	Line up color palette in order of brightness by the quicksort method.  

	If topDown is true, arranges in descending order of brightness. If false, in ascending order.
*********************************************************************************************************/
void
CColorPalette::QuickSortTopDown(																				
				bool					topDown						// true: descending order false: ascending order
)
{
	CData	yChannel;
	u32*	pIndex;

	u32 number = GetNumber();

	yChannel.Create( number, 0 );
	pIndex = new u32[ number ];

	u8 red, green ,blue, y;
	u32 i;

	// Calculate brightness
	for( i=0; i<number; i++ )
	{
		m_redChannel.  GetValue( &red, i  );
		m_greenChannel.GetValue( &green, i  );
		m_blueChannel. GetValue( &blue, i  );
		y = TLB_GET_Y( red, green, blue );
		yChannel.SetValue( y, i );
		
		pIndex[i] = ( red << 16 ) | ( green << 8 ) | blue;
	}

	// Sort by brightness
	yChannel.QuickSortTopDown( pIndex, 0, number-1, topDown );

	// Sort Colors
	for( i=0; i<GetNumber(); i++ )
	{
		red		= ( pIndex[i] >> 16 ) & 0xff;
		green	= ( pIndex[i] >> 8 ) & 0xff;
		blue	=	pIndex[i] & 0xff;
		SetValue( red, green, blue, i );
	}

	delete[] pIndex;
}


/*********************************************************************************************************
	TXLib::CColorPalette::CommonLowerBit

	Convert into bit value indecated by pseudo-argument by standardizing lower bit of 8 bit color value.   
*********************************************************************************************************/
void
CColorPalette::CommonLowerBit(
				u8							virtualBit							// Pseudo-converted bit value
)
{
	assert( virtualBit <= 8 );

	u8 restBit = 8 - virtualBit;

	for( u32 i=0; i<GetNumber(); i++ )
	{
		u8 red, green ,blue;
		u8 redBuf, greenBuf, blueBuf;

		GetValue( &red, &green, &blue, i );

		redBuf		=  red		>> restBit;
		greenBuf	=  green	>> restBit;
		blueBuf		=  blue		>> restBit;

		//			Upper 5 bit				Lower 3 bit
		red		= (redBuf	<< restBit)	| (redBuf	>> ( virtualBit - restBit));
		green	= (greenBuf << restBit)	| (greenBuf	>> ( virtualBit - restBit));
		blue	= (blueBuf	<< restBit)	| (blueBuf	>> ( virtualBit - restBit));

		SetValue(red, green, blue, i );
	}
}


} // namespace
