/****************************************************************************
  Project: NITRO-SampleTools - TXLib
  File:    TXLib_Texture.cpp

  Copyright 2004,2005 Nintendo Co.,Ltd.	All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  Handles Texture Class

  $Log: TXLib_Texture.cpp,v $
  Revision 1.8  2005/06/20 06:54:11  mominoki
  Changed copyright. Made revision adjustments. 

  Revision 1.2	2004/10/27		mominoki
  In ChangeDirectToTex4x4(), changed m_colorPaletteAddress from 2 byte to 4 byte.

  Revision 1.1	2004/04/15		mominoki
  Revised comments.

  Revision 1.0	2004/03/03		mominoki
  Initial release.

  $NoKeywords: $
 ***************************************************************************/
#define TLB_EXPORT_

#include <cstring>
#include <TXLib/Base/TXLib_Texture.h>
#include <TXLib_Define.h> 

namespace TXLib{

// Global variable
CDefaultColorProcessor		gDefaultColorProcessor;


/*********************************************************************************************************
	TXLib::CTexture::CTexture

	Constructor
*********************************************************************************************************/
CTexture::CTexture()
:
	m_colorIndexImage(),
	m_colorPalette(),
	m_directColorImage(),
	m_alphaImage(),
	m_colorPaletteAttribute(),
	m_colorPaletteAddress(),
	m_type(kType_not),
	m_pColorProcessor( &gDefaultColorProcessor )
{
}

/*********************************************************************************************************
	TXLib::CTexture::CTexture

	Copy constructor
*********************************************************************************************************/
CTexture::CTexture(
	const	CTexture&		rTexture			// CTexture pointer to copy source
) :
	m_colorIndexImage		(rTexture.m_colorIndexImage	),
	m_colorPalette			(rTexture.m_colorPalette	),
	m_directColorImage		(rTexture.m_directColorImage),
	m_alphaImage			(rTexture.m_alphaImage		),
	m_colorPaletteAttribute	(rTexture.m_colorPaletteAttribute),
	m_colorPaletteAddress	(rTexture.m_colorPaletteAddress),
	m_type					(rTexture.m_type			),
	m_pColorProcessor		( rTexture.m_pColorProcessor )
{

}

/*********************************************************************************************************
	TXLib::CTexture::operator=

	Assignemnt operator
*********************************************************************************************************/
CTexture&
CTexture::operator=(
	const	CTexture&		rTexture			// CTexture pointer to copy source
)
{
	if (&rTexture != this)
	{
		m_colorIndexImage		= rTexture.m_colorIndexImage;
		m_colorPalette			= rTexture.m_colorPalette;
		m_directColorImage		= rTexture.m_directColorImage;
		m_alphaImage			= rTexture.m_alphaImage;
		m_colorPaletteAttribute	= rTexture.m_colorPaletteAttribute;
		m_colorPaletteAddress	= rTexture.m_colorPaletteAddress;
		m_type					= rTexture.m_type;
		m_pColorProcessor		= rTexture.m_pColorProcessor;
	}
	return *this;
}

/*********************************************************************************************************
	TXLib::CTexture::~CTexture

	Destructor
*********************************************************************************************************/
CTexture::~CTexture()
{ 
	Clear(); // Clear
}

/*********************************************************************************************************
	TXLib::CTexture::GetTex4x4Palette
	
	Gets a tex4x4 palette from color memory.
	IS called from ChangeTex4x4ToDirect().
*********************************************************************************************************/
bool
CTexture::GetTex4x4Palette( 
			CColorPalette*			pDstColorPalette,				// Color memory
	const	CColorPalette*			pSrcColorPalette,				// Color memory
			u8						attribute,						// Palette attribute
			u16						paletteAddress					// Palette address
)
{
	assert( ( pDstColorPalette != NULL ) && ( pSrcColorPalette != NULL ) );

	u8 red, green, blue;
	u8 red0, green0, blue0;
	u8 red1, green1, blue1;

	switch( attribute )
	{
		case 0:		// 3 color, no interpolation 
			for( u32 i=0; i<3; i++ )
			{
				pSrcColorPalette->GetValue( &red, &green, &blue, ( paletteAddress * 2 + i)	);		// Get
				pDstColorPalette->SetValue( red, green, blue, i  );
			}
			break;
		case 1:		// 3 color, no interpolation 
			pSrcColorPalette->GetValue( &red0, &green0, &blue0, ( paletteAddress * 2 + 0)  );		// Get
			pDstColorPalette->SetValue( red0, green0, blue0 , 0 );

			pSrcColorPalette->GetValue( &red1, &green1, &blue1, ( paletteAddress * 2 + 1)  );		// Get
			pDstColorPalette->SetValue(  red1, green1, blue1, 1 );

			red		= TLB_GET_TEX4X4LINEARPALETTE_COLOR2( red0,		red1	);
			green	= TLB_GET_TEX4X4LINEARPALETTE_COLOR2( green0,	green1	);
			blue	= TLB_GET_TEX4X4LINEARPALETTE_COLOR2( blue0,	blue1	);

			pDstColorPalette->SetValue( red, green, blue, 2  );
			break;
		case 2:		//  3 color, no interpolation 
			for( u32 i=0; i<4; i++ )
			{
				pSrcColorPalette->GetValue( &red, &green, &blue, ( paletteAddress * 2 + i)	);		// Get
				pDstColorPalette->SetValue( red, green, blue , i );
			}
			break;
		case 3:		//  4 color, with interpolation
			pSrcColorPalette->GetValue( &red0, &green0, &blue0, ( paletteAddress * 2 + 0) );		// Get
			pDstColorPalette->SetValue( red0, green0, blue0 , 0 );

			pSrcColorPalette->GetValue(  &red1, &green1, &blue1, ( paletteAddress * 2 + 1));		// Get
			pDstColorPalette->SetValue( red1, green1, blue1, 1	);

			red		= TLB_GET_TEX4X4LINEARPALETTE_COLOR2_FOUR( red0,	red1	);
			green	= TLB_GET_TEX4X4LINEARPALETTE_COLOR2_FOUR( green0,	green1	);
			blue	= TLB_GET_TEX4X4LINEARPALETTE_COLOR2_FOUR( blue0,	blue1	);

			pDstColorPalette->SetValue( red, green, blue, 2  );

			red		= TLB_GET_TEX4X4LINEARPALETTE_COLOR3_FOUR( red0,	red1	);
			green	= TLB_GET_TEX4X4LINEARPALETTE_COLOR3_FOUR( green0,	green1	);
			blue	= TLB_GET_TEX4X4LINEARPALETTE_COLOR3_FOUR( blue0,	blue1	);

			pDstColorPalette->SetValue(  red, green, blue, 3 );
			break;
		default:
			return false;
	}

	return true;
}

/*********************************************************************************************************
	TXLib::CTexture::ChangePaletteTypeToDirect

	Converts the palette format to direct.

	* If the palette format contains an alpha image, this will copy the alpha image.

	* If there is no alpha image this will set the alpha image alpha value to default to 0xff (opaque).
*********************************************************************************************************/
bool
CTexture::ChangePaletteTypeToDirect()
{
	m_directColorImage.Create( GetWidth(), GetHeight(), 0,0,0);

	// Create alpha image
	if( !m_alphaImage.IsAvailable() )
	{
		m_alphaImage.Create( GetWidth(), GetHeight(), 0xff );
	}

	// Create direct color image
	for (u16 y=0; y<GetHeight(); y++)
	{
		for (u16 x=0; x<GetWidth(); x++)
		{
			u8	red, green, blue;
			u8	index;

			m_colorIndexImage.GetValue( &index, x, y );
			m_colorPalette.GetValue( &red, &green, &blue, index );
			m_directColorImage.SetValue( red,  green,  blue, x, y  );
		}
	}
	m_colorIndexImage.Clear();
	m_colorPalette	 .Clear();

	m_type = kType_direct;

	return true;
}

/*********************************************************************************************************
	TXLib::CTexture::ChangeDirectToPaletteType

	Converts from direct to palette format.
*********************************************************************************************************/
bool
CTexture::ChangeDirectToPaletteType(
			tType		type		// Texture format
)
{
	int paletteSize;

	switch (type)
	{
		case kType_palette4 :	paletteSize =	4;	break;
		case kType_a5i3:		paletteSize =	8;	break;
		case kType_palette16:	paletteSize =  16;	break;
		case kType_a3i5		:	paletteSize =  32;	break;
		case kType_palette256 : paletteSize = 256;	break;
		default: assert(false);
	}
	m_colorPalette.   Create( paletteSize, 0,0,0);
	m_colorIndexImage.Create( GetWidth(), GetHeight() );

	// Create palette, color index 
	if( !m_pColorProcessor->ConvertToPaletteType(this)) return false;

	m_type = type;
	m_directColorImage.Clear();

	// When there is no alpha, clear alpha 
	if( (type != kType_a5i3) && (type != kType_a3i5) )
	{
		m_alphaImage.Clear();
	}

	return true;
}

/*********************************************************************************************************
	TXLib::CTexture::ChangePaletteType

	Converts a palette fromat to another palette format.
*********************************************************************************************************/
bool
CTexture::ChangePaletteType(
		tType		type
)
{
	u32	paletteSize;

	switch (type)
	{
		case kType_palette4 :	paletteSize =	4;	break;
		case kType_a5i3:		paletteSize =	8;	break;
		case kType_palette16 :	paletteSize =  16;	break;
		case kType_a3i5:		paletteSize =	32;	break;
		case kType_palette256 : paletteSize = 256;	break;
		default: assert(false);
	}

	// Palette, color index 
	if (m_colorPalette.GetNumber() <= paletteSize)
	{
		m_type = type;
	}
	else if (m_colorPalette.GetNumber() > paletteSize)
	{
		ChangePaletteTypeToDirect();
		ChangeDirectToPaletteType(type);
	}

	// Alpha 
	if( (type != kType_a5i3) && (type != kType_a3i5) )
	{
		m_alphaImage.Clear();
	}
	else
	{
		if( !m_alphaImage.IsAvailable() )
		{
			m_alphaImage.Create( GetWidth(), GetHeight(), 0xff );
		}
	}

	return true;
}

/*********************************************************************************************************
	TXLib::CTexture::ChangeTex4x4ToDirect

	Converts tex4x4 to direct
*********************************************************************************************************/
bool
CTexture::ChangeTex4x4ToDirect()
{
	u32 width  = m_colorIndexImage.GetWidth();
	u32 height = m_colorIndexImage.GetHeight();

	m_directColorImage.Create(width, height, 0,0,0);

	u32	block = 0;	// Number of block

	for (u32 y=0; y<height; y+=TLB_4)				// row
	{
		for (u32 x=0; x<width; x+=TLB_4, block++)	// column
		{
			CColorPalette	fourColor;		// tex4x4 palette
			fourColor.Create( 4, 0, 0, 0 );

			u32		paletteAddress;
			u8		paletteAttribute;

			m_colorPaletteAddress  .GetValue( &paletteAddress, block   );
			m_colorPaletteAttribute.GetValue( &paletteAttribute, block );

			// Get the tex4x4 palette
			if(!GetTex4x4Palette(
								&fourColor,
								&m_colorPalette,
								paletteAttribute,
								paletteAddress )) return false;

			// Assign to direct color image
			for (u32 block_y=0; block_y<TLB_4; block_y++)
			{
				for( u32 block_x=0; block_x<TLB_4; block_x++ )
				{
					u32 posx = x + block_x;
					u32 posy = y + block_y;

					u8	index;
					m_colorIndexImage.GetValue( &index, posx, posy );

					u8 red, green, blue;

					fourColor.GetValue( &red, &green, &blue, index );
					m_directColorImage.SetValue( red, green, blue, posx, posy  );
				}
			}
		}
	}

	m_colorPaletteAddress.	Clear();
	m_colorPaletteAttribute.Clear();
	m_colorPalette.			Clear();
	m_colorIndexImage.		Clear();

	m_type = kType_direct;

	return true;
}

/*********************************************************************************************************
	TXLib::CTexture::ChangeDirectToTex4x4

	Convert direct to tex4x4
*********************************************************************************************************/
bool
CTexture::ChangeDirectToTex4x4()
{
	u16 width  = m_directColorImage.GetWidth();
	u16 height = m_directColorImage.GetHeight();

	m_colorIndexImage.		Create( width, height );
	m_colorPalette.			Create( ( width * height ) / ( TLB_4 * TLB_4 ) * 4, 0, 0, 0 );
	m_colorPaletteAttribute.Create( ( width * height ) / ( TLB_4 * TLB_4), 0 );
	m_colorPaletteAddress.	Create( 4, ( width * height ) / ( TLB_4 * TLB_4), 0 );

	// Create tex4x4
	if( !m_pColorProcessor->ConvertToTex4x4(this)) return false;

	m_directColorImage.Clear();
	m_alphaImage.Clear();
	m_type = kType_tex4x4;

	return true;
}

/*********************************************************************************************************
		public
*********************************************************************************************************/
/*********************************************************************************************************
	TXLib::CTexture::~Clear

	Clear memory
*********************************************************************************************************/
void
CTexture::Clear()
{
	m_colorIndexImage		.Clear();
	m_colorPalette			.Clear();
	m_directColorImage		.Clear();
	m_alphaImage			.Clear();
	m_colorPaletteAttribute	.Clear();
	m_colorPaletteAddress	.Clear();

	m_type = kType_not;
	m_pColorProcessor		= &gDefaultColorProcessor;
}

/*********************************************************************************************************
	TXLib::CTexture::GetWidth

	Gets image size, width

    If it has a color index image, the image size width will be the width in the color index size.
	If it has a direct color image, the image size width will be the direct color image width.
	If it has neither, width=0.
*********************************************************************************************************/
inline u16
CTexture::GetWidth() const
{
	u16 width = 0;

	if( m_colorIndexImage.IsAvailable() )
	{
		width = m_colorIndexImage.GetWidth();		// Color index width
	}
	else if( m_directColorImage.IsAvailable() )
	{
		width = m_directColorImage.GetWidth();		// Direct color index width

	}
	return width;
}

/*********************************************************************************************************
	TXLib::CTexture::GetHeight

	Gets image size height.

	Refer to GetWidth().
*********************************************************************************************************/
inline u16
CTexture::GetHeight() const
{
	u16 height = 0;

	if( m_colorIndexImage.IsAvailable() )
	{
		height = m_colorIndexImage.GetHeight();			// Color index height
	}
	else if( m_directColorImage.IsAvailable() )
	{
		height = m_directColorImage.GetHeight();		// Direct color image height
	}
	return height;
}

/*********************************************************************************************************
	TXLib::CTexture::SetColorProcessor

	Set a color processor object in CTexture

	If NULL is specified, the CDefaultColorProcessor (default) will be set.
*********************************************************************************************************/
inline void
CTexture::SetColorProcessor(
				CColorProcessor*		pColorProcessor			// Color process object
)
{
	m_pColorProcessor = (pColorProcessor == NULL) ? &gDefaultColorProcessor : pColorProcessor;
}

/*********************************************************************************************************
	TXLib::CTexture::Convert

	Converts from other texture formats
*********************************************************************************************************/
bool
CTexture::Convert(
			tType		convertType				// Type of texture format to convert
)
{
	switch ( convertType )
	{
		case kType_palette4:
		case kType_a5i3:
		case kType_palette16:
		case kType_a3i5:
		case kType_palette256:
			if (IsPaletteType())
			{
				if( m_type != convertType )
				{
					ChangePaletteType(convertType);			// palette type -> palette type
				}
			}
			else if (m_type == kType_direct)
			{
				ChangeDirectToPaletteType(convertType);		// direct -> palette type
			}
			else if (m_type == kType_tex4x4)
			{
				ChangeTex4x4ToDirect();						// tex4x4 -> direct
				ChangeDirectToPaletteType(convertType);		// direct -> palette type
			}
			break;
		case kType_tex4x4:
			if((GetWidth() % 4 ) || (GetHeight() % 4 ) ) return false;	// Size check (Are width and height multiples of 4?)

			if (IsPaletteType())
			{
				ChangePaletteTypeToDirect();				// palette type -> direct
				if(!ChangeDirectToTex4x4()) return false;	// direct -> tex4x4
			}
			else if (m_type == kType_direct)
			{
				if(!ChangeDirectToTex4x4()) return false;	// direct -> tex4x4
			}
			break;
		case kType_direct:
			if (IsPaletteType())
			{
				ChangePaletteTypeToDirect();				// palette type -> direct
			}
			else if (m_type == kType_tex4x4)
			{
				ChangeTex4x4ToDirect();						// tex4x4 -> direct
			}
			break;
		default:
			return false;
	}
	return true;
}

/*********************************************************************************************************
	TXLib::CTexture::IsPaletteType

	The texture format determines the palette format.
*********************************************************************************************************/
bool
CTexture::IsPaletteType() const
{
	bool result;

	switch( m_type )
	{
		case kType_palette4 :
		case kType_a5i3:
		case kType_palette16 :
		case kType_a3i5:
		case kType_palette256 :
			result = true;
			break;
		default:
			result = false;
			break;
	}

	return result;
}

/*********************************************************************************************************
	TXLib::CTexture::AlignmentImageSize

	* Aligns the texture size to the specified size (enlarge).
	* The enlarged area becomes the clamp image.
	* The specified size must be larger than the source texture size.
*********************************************************************************************************/
bool
CTexture::AlignmentImageSize(
			u16					width,							// size width
			u16					height							// size height
)
{
	assert( (width >0) && ( height>0));

	u16 oldWidth	= GetWidth();
	u16 oldHeight	= GetHeight();

	u8	red, green, blue, alpha;
	u8	index;

	// Adjust direct color image
	if( m_directColorImage.IsAvailable() )
	{
		CDirectColorImage	directColorImage;
		directColorImage.Create( width, height, 0, 0, 0 );

		for( u16 y=0; y<oldHeight; y++)
		{
			for( u16 x=0; x<oldWidth; x++)
			{
				m_directColorImage.GetValue(&red, &green, &blue, x, y );
				directColorImage.SetValue( red, green ,blue, x, y );
			}
			// Expansion area
			for( u16 x=oldWidth; x<width; x++)
			{
				directColorImage.SetValue( red, green ,blue, x, y );
			}
		}
		// Expansion area
		for( u16 y=oldHeight; y<height; y++)
		{
			for( u16 x=0; x<width; x++)
			{
				directColorImage.GetValue(&red, &green, &blue, x, oldHeight-1 );
				directColorImage.SetValue( red, green ,blue, x, y );
			}
		}
		m_directColorImage = directColorImage;
	}

	// Adjust alpha image
	if( m_alphaImage.IsAvailable() )
	{
		CAlphaImage		alphaImage;
		alphaImage.Create( width, height, 0xff );

		for( u16 y=0; y<oldHeight; y++)
		{
			for( u16 x=0; x<oldWidth; x++)
			{
				m_alphaImage.GetValue(&alpha, x, y );
				alphaImage.SetValue( alpha, x, y );
			}
			// Expansion area
			for( u16 x=oldWidth; x<width; x++)
			{
				alphaImage.SetValue( alpha, x, y );
			}
		}
		// Expansion area
		for( u16 y=oldHeight; y<height; y++)
		{
			for( u16 x=0; x<width; x++)
			{
				alphaImage.GetValue( &alpha, x, oldHeight-1 );
				alphaImage.SetValue( alpha, x, y );
			}
		}
		m_alphaImage = alphaImage;
	}

	// Adjust color index image
	if( m_colorIndexImage.IsAvailable() )
	{
		CColorIndexImage		colorIndexImage;
		colorIndexImage.Create( width, height );

		for( u16 y=0; y<oldHeight; y++)
		{
			for( u16 x=0; x<oldWidth; x++)
			{
				m_colorIndexImage.GetValue(&index, x, y );
				colorIndexImage.SetValue( index, x, y );
			}
			// Expansion area
			for( u16 x=oldWidth; x<width; x++)
			{
				colorIndexImage.SetValue( index, x, y );
			}
		}
		// Expansion area
		for( u16 y=oldHeight; y<height; y++)
		{
			for( u16 x=0; x<width; x++)
			{
				colorIndexImage.GetValue( &index, x, oldHeight-1 );
				colorIndexImage.SetValue( index, x, y );
			}
		}
		m_colorIndexImage = colorIndexImage;
	}

	// Adjust palette attribute
	if( m_colorPaletteAttribute.IsAvailable() )
	{
		CData		paletteAttribute;
		paletteAttribute.Create( ( width * height ) / ( TLB_4 * TLB_4), 0 );

		u8 attribute;
		u32 block=0;
		u32 newBlock=0;
		for (u16 y=0; y<oldHeight; y+=TLB_4)				// row
		{
			for (u16 x=0; x<width; x+=TLB_4, newBlock++ )	// column
			{
				if( x<oldWidth)
				{
					m_colorPaletteAttribute.GetValue( &attribute, block );
					paletteAttribute.SetValue( attribute, newBlock );
					block++;
				}
				else
				{
					// Expansion area
					paletteAttribute.SetValue( attribute, newBlock );
				}
			}
		}
		// Expansion area
		u16 blockWidth = width / TLB_4;
		for (u16 y=oldHeight; y<height; y+=TLB_4)			// row
		{
			for (u16 x=0; x<width; x+=TLB_4, newBlock++ )	// column
			{
				paletteAttribute.GetValue( &attribute, newBlock-blockWidth );
				paletteAttribute.SetValue( attribute, newBlock );
			}
		}
		m_colorPaletteAttribute = paletteAttribute;
	}

	// Adjust palette address
	if( m_colorPaletteAddress.IsAvailable() )
	{
		CByteData	paletteAddress;
		paletteAddress.Create( 2,  ( width * height ) / ( TLB_4 * TLB_4), 0 );

		u32 address;
		u32 block=0;
		u32 newBlock=0;
		for (u16 y=0; y<oldHeight; y+=TLB_4)				// row
		{
			for (u16 x=0; x<width; x+=TLB_4, newBlock++ )	// column
			{
				if( x<oldWidth)
				{
					m_colorPaletteAddress.GetValue( &address, block );
					paletteAddress.SetValue( address, newBlock );
					block++;
				}
				else
				{
					// Expansion area
					paletteAddress.SetValue( address, newBlock );
				}
			}
		}
		// Expansion area
		u16 blockWidth = width / TLB_4;
		for (u16 y=oldHeight; y<height; y+=TLB_4)				// row
		{
			for (u16 x=0; x<width; x+=TLB_4, newBlock++ )	// column
			{
				paletteAddress.GetValue( &address, newBlock-blockWidth );
				paletteAddress.SetValue( address, newBlock );
			}
		}
		m_colorPaletteAddress = paletteAddress;
	}

	return true;
}

} // namespace
