/****************************************************************************
  Project: NITRO-SampleTools - TXLib
  File:    TXLib_DefaultColorProcessor.cpp

  Copyright 2004,2005 Nintendo Co.,Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  Class that handles Color processing.  Called from TXLib::CTexture.
  This class is called in the initial state of CTexture.
  To call another TXLib_ColorProcessor, prepare something like this class
  independently.

  $Log: TXLib_DefaultColorProcessor.cpp,v $
  Revision 1.7  2005/06/20 06:54:11  mominoki
  Changed copyright. Made revision adjustments. 

  Revision 1.1  2004/04/15		mominoki
  Corrected comment

  Revision 1.0  2004/03/03		mominoki
  Initial registration

  $NoKeywords: $
 ***************************************************************************/
#define TLB_EXPORT_

#include <cstdlib>
#include <new.h>
#include <cstring>
#include <cassert>
#include <TXLib_Define.h> 
#include <TXLib/Convert/TXLib_DefaultColorProcessor.h>

namespace std{ using :: memset; }

namespace TXLib{
/*********************************************************************************************************
	TXLib::CDefaultColorProcessor::CDefaultColorProcessor

	Constructor
*********************************************************************************************************/
CDefaultColorProcessor::CDefaultColorProcessor()
{
		std::memset( (void*)&m_convertFactor, 0, sizeof(tConvertFactor));
};

// virtual
/*********************************************************************************************************
	TXLib::CDefaultColorProcessor::FindNearColor

	Searches for a near color.

	If you want to avoid a particular index (when the transparent color is index 3, etc), specify the index in avoidNumber.
	In this case when it searches it will not select this index. When avoidNumber=-1, it will be disabled.
*********************************************************************************************************/
int	CDefaultColorProcessor::FindNearColor(
	const	CColorPalette*				pPalette,					// Palette to be searched
			u8							red,
			u8							green,
			u8							blue,
			s16							avoidNumber					// Index number you want to avoid, such as transparent color.
)
{
	assert( pPalette != NULL );
	assert( pPalette->IsAvailable() );

	u8	index				= 0;					// Color index
	u32 distanceColorMin	= 0xffffffff;			// Initial value Max
	u8 paletteRed, paletteGreen, paletteBlue;

	for(u32  i=0; i<pPalette->GetNumber(); i++)
	{
		if( i == avoidNumber ) continue;			// Will not select a particular index

		pPalette->GetValue( &paletteRed, &paletteGreen, &paletteBlue, i );					// Get color value

		int distanceWeightRed	= TLB_WEIGHT_RED	* ( red		- paletteRed );
		int distanceWeightGreen	= TLB_WEIGHT_GREEN	* ( green	- paletteGreen );
		int distanceWeightBlue	= TLB_WEIGHT_BLUE	* ( blue	- paletteBlue );

		u32	distanceColor = distanceWeightRed * distanceWeightRed  +
							distanceWeightGreen * distanceWeightGreen +
							distanceWeightBlue * distanceWeightBlue;	// Seek color distance with least squares method

		if( distanceColor < distanceColorMin )						// The one with the shortest color distance will be a color index value.
		{
			distanceColorMin = distanceColor;
			index = i;												// Changes the index value
			if( distanceColorMin == 0 ) break;						// Exit the loop if all the color values are the same.
		}
	}

	return index;
}

/*********************************************************************************************************
	TXLib::CDefaultColorProcessor::ConvertToPaletteType

	Converts from direct to palette format texture

	* Called from CTexture::ChangeDirectToPaletteType()
	* When m_convertFactor.transparencyFlg is true, it will create a texture with the palette color 0 as the transparent color.
	* When m_convertFactor.alphaColorCutFlg is true, for the above it will make a transparent color from the alpha image, and allocate it to the palette color 0.
   Texels with an alpha value of less than 0x80 will be taken as transparent color.
	* When m_convertFactor.pUsePalette is specified, it will create a texture with that palette. In that case, it will not carry out the color reduction process.
*********************************************************************************************************/
bool
CDefaultColorProcessor::ConvertToPaletteType(
			CTexture*					pTexture					// Texture
)
{
	assert( pTexture );
	assert( pTexture->GetColorIndexImage().IsAvailable() && pTexture->GetColorPalette().IsAvailable() && 
			pTexture->GetDirectColorImage().IsAvailable());

	u16 width	= pTexture->GetWidth();
	u16 height	= pTexture->GetHeight();

	// Create color memory
	CColorPalette	colorMemory;
	colorMemory.Create( width * height, 0, 0, 0 );

	// Create color memeory from direct color image
	if( m_convertFactor.transparencyFlg )
	{
		//--------------------------------------------------------------
		//	When using transparent color (or an alpha image), this sets so that the transparent
		// color is not included in the color reduction algorithm.
		//--------------------------------------------------------------
		if( !m_convertFactor.alphaColorCutFlg )
		{
			//--------------------------------------------------------------
			//	When there is a transparent color, create an alpha image
			//--------------------------------------------------------------
			tTLBRGB		transparentColor;

			if( m_convertFactor.pUsePalette != NULL)
			{
				// When an palette is specified, this sets the palette's color 0 as the transparent color.
				m_convertFactor.pUsePalette->GetValue(	&transparentColor.red,
														&transparentColor.green,
														&transparentColor.blue,
														0);
			}
			else
			{
				transparentColor.red	= m_convertFactor.transparentColor.red;
				transparentColor.green	= m_convertFactor.transparentColor.green;
				transparentColor.blue	= m_convertFactor.transparentColor.blue;
			}

			// This creates a binary alpha image (0x00 or 0xff) from a direct color image and a transparent color.
			u8	red, green, blue;

			for( u16 y=0; y<height; y++)
			{
				for( u16 x=0; x<width; x++)
				{
					pTexture->GetDirectColorImage().GetValue( &red, &green, &blue, x, y );

					if( (transparentColor.red == red ) && (transparentColor.green == green ) &&
						(transparentColor.blue == blue ) )
					{
						// If it is a transparent color, set the alpha value to 0
						pTexture->GetAlphaImage().SetValue( 0, x, y );
					}
					else
					{
						// If it is not a transparent color, set the alpha value to 0xff
						pTexture->GetAlphaImage().SetValue( 0xff, x, y );
					}
				}
			}
		}

		// Copies a direct color image to color memory (colors with alpha values less than 0x80 will be cut)
		CreateColorMemory( &colorMemory, &pTexture->GetDirectColorImage(), &pTexture->GetAlphaImage(), 0x80 );
	}
	else
	{
		// When no transparent color is used, this copies the direct color image to color memory, unchanged.
		CreateColorMemory( &colorMemory, &pTexture->GetDirectColorImage(), NULL, 0x00 );
	}

	u32 nObjectiveColor = pTexture->GetColorPalette().GetNumber();		// Targeted color reduction number
	s16 avoidNumber = -1;	// Specify so you can search color indexes in all indexes.										// Transparent color's index.

	if( m_convertFactor.pUsePalette != NULL )
	{
		//--------------------------------------------------------------
		//	Use the palettes that are set in the conversion factor
		//--------------------------------------------------------------
		if( !m_convertFactor.pUsePalette->IsAvailable() ) return false;

		pTexture->GetColorPalette() = *m_convertFactor.pUsePalette; // Use the conversion factor's palettes

		if(	m_convertFactor.pUsePalette->GetNumber() > nObjectiveColor )
		{
			// If too many palettes are specified, the excess palettes will be cut.
			pTexture->GetColorPalette().ChangeNumber( nObjectiveColor );
		}
	}
	else
	{
		//--------------------------------------------------------------
		//	Reduce color and create palettes
		//--------------------------------------------------------------
		CColorPalette	colorMemoryBuf;

		colorMemory.CommonLowerBit(5);										// Pseudo convert 8-bit color values to 5-bit
		colorMemoryBuf = colorMemory;										// Copy to buffer

		colorMemoryBuf.CutCommonColor();									// Cut common color

		if( colorMemoryBuf.GetNumber() <= nObjectiveColor )
		{
			pTexture->GetColorPalette() = colorMemoryBuf;					// Copy as is
		}
		else
		{
			CompressPalette_MedhianCut( &pTexture->GetColorPalette(), &colorMemory  );	// Compress using median cut
			pTexture->GetColorPalette().CutCommonColor();								// Cut common color
		}

		//	Create palette color 0
		if( m_convertFactor.transparencyFlg )
		{
			u32 nColor = pTexture->GetColorPalette().GetNumber();
			
			// If it is possible to add a palette, add one.
			if( nColor < nObjectiveColor )
			{
				nColor += 1;
				pTexture->GetColorPalette().ChangeNumber( nColor );
			}

			// Shift palettes by one. End will be overwritten, and therefore cut.
			for( u16 i=0; i<(nColor - 1); i++ )
			{
				u8 red, green, blue;

				pTexture->GetColorPalette().GetValue( &red, &green, &blue, (nColor - 2 - i) );
				pTexture->GetColorPalette().SetValue( red, green, blue, (nColor -1 - i) );
			}

			// Set the palette color 0 to be transparent color
			pTexture->GetColorPalette().SetValue(	m_convertFactor.transparentColor.red,
													m_convertFactor.transparentColor.green,
													m_convertFactor.transparentColor.blue,
													0);

			avoidNumber = 0;	// Set to avoid "0" because palette color 0 is the transparent color.
		}
	}

	//--------------------------------------------------------------
	//	Create color index. For alpha values less than 0x80, set as transparent colors with index "0"
	//--------------------------------------------------------------
	for( u16 y = 0; y<height; y++)
	{
		for( u16 x=0; x<width; x++)
		{
			int index;
			u8 alpha;

			pTexture->GetAlphaImage().GetValue( &alpha, x, y );

			if( m_convertFactor.transparencyFlg & (alpha < 0x80) )
			{
				index = 0;			// Transparent color
			}
			else
			{
				u8 red, green, blue;

				pTexture->GetDirectColorImage().GetValue( &red, &green, &blue, x, y  );					// Get image color value 
				index = FindNearColor( &pTexture->GetColorPalette(), red, green, blue, avoidNumber );	// Search color index from palette
				if( index == -1) return false;
			}

			pTexture->GetColorIndexImage().SetValue( (u8)index, x, y );		// Set index value
		}
	}

	return true;
}

/*********************************************************************************************************
	TXLib::CDefaultColorProcessor::ConvertToTex4x4

	Convert from direct to tex4x4
	Called by CTexture::ChangeDirectToTex4x4()
*********************************************************************************************************/
bool
CDefaultColorProcessor::ConvertToTex4x4(
			CTexture*					pTexture					// Texutre
)
{
	assert( pTexture );

	CAlphaImage*			pAlphaImage				= &pTexture->GetAlphaImage();
	CByteData*				pColorPaletteAddress	= &pTexture->GetColorPaletteAddress();			
	CColorIndexImage*		pColorIndexImage		= &pTexture->GetColorIndexImage();
	CColorPalette*			pColorPalette			= &pTexture->GetColorPalette();
	CData*					pColorPaletteAttribute	= &pTexture->GetColorPaletteAttribute();
	CDirectColorImage*		pDirectColorImage		= &pTexture->GetDirectColorImage();

	assert( pColorIndexImage->IsAvailable()		&& pColorPalette->IsAvailable() && 
			pDirectColorImage->IsAvailable()	&& pAlphaImage->IsAvailable() &&
			pColorPaletteAddress->IsAvailable() && pColorPaletteAttribute->IsAvailable());

	u16 width  = pTexture->GetWidth();
	u16 height = pTexture->GetHeight();

	u32	block			= 0;	// Number of block
	u32 nNewColor		= 0;	// Palette total color count

	for (u16 y=0; y<height; y+=TLB_4)				// row
	{
		for (u16 x=0; x<width; x+=TLB_4, block++)	// column
		{
			//--------------------------------------------------------------
			//	Create a one block direct color image.
			//--------------------------------------------------------------
			CDirectColorImage	blockDirectColorImage;
			blockDirectColorImage.Create( TLB_4, TLB_4, 0, 0, 0 );		// Create one block image

			MakeBlockDirectColorImage_Tex4x4(
						&blockDirectColorImage,
						pDirectColorImage,
						x,
						y);

			//--------------------------------------------------------------
			//	Create one block alpha image
			//--------------------------------------------------------------
			CAlphaImage		blockAlphaImage;
			blockAlphaImage.Create( TLB_4, TLB_4, 0xff );			// Default is opaque

			bool transparencyFlg = m_convertFactor.transparencyFlg;		// Transparent color mode flag. If there are no transparent colors in the block, always use in 4 color mode.

			if( transparencyFlg )
			{
				if( m_convertFactor.alphaColorCutFlg )
				{
					// Create alpha with alpha. If alpha values are less than 0x80, set alpha values to 0.
					transparencyFlg = MakeBlockAlphaImageUseAlpha_Tex4x4(
							&blockAlphaImage,
							pAlphaImage,
							x,
							y,
							0x80);
				}
				else
				{
					// Create alpha with transparent color. Set alpha to 0 if the image color value is the same as the transparent color.
					transparencyFlg = MakeBlockAlphaImageUseTransparentColor_Tex4x4(
							&blockAlphaImage,
							&blockDirectColorImage,
							m_convertFactor.transparentColor);
				}
			}

			//--------------------------------------------------------------
			//	Get a tex4x4 palette (0-4 color) from one block.
			//--------------------------------------------------------------
			CColorPalette	fourColorPalette;

			fourColorPalette.Create( 4, 0, 0, 0 );

			if( !Get4Color_Tex4x4(
						&fourColorPalette,
						transparencyFlg,
						&blockDirectColorImage,
						&blockAlphaImage,
						m_convertFactor.algorithm
						)) return false;

			//--------------------------------------------------------------
			//	Create color palette attribute
			//--------------------------------------------------------------
			u8 attribute = 0;
			
			if( fourColorPalette.GetNumber() < 3 )
			{
				attribute |= 1;			// Linear interpolation
			}

			if( !transparencyFlg )
			{
				attribute |= 2;			// 4 colors
			}

			pColorPaletteAttribute->SetValue( attribute, block  );


			//--------------------------------------------------------------
			//	Create color palette and color palette address
			//--------------------------------------------------------------
			u8	nCompareColor	= fourColorPalette.GetNumber();	// Number of colors in tex4x4 palette

			if( nCompareColor == 0 )
			{
				// If all are transparent color, palette address is "0"
				pColorPaletteAddress->SetValue( 0, block );
			}
			else
			{
				// Search for the same tex4x4 palette
				int	paletteIndex = -1;

				if( GetConvertFactor().compressFlg )
				{
					paletteIndex = FindNearPalette(
								&fourColorPalette,
								pColorPalette,
								nNewColor,
								m_convertFactor.commonColorRange.red	<< 3,
								m_convertFactor.commonColorRange.green	<< 3,
								m_convertFactor.commonColorRange.blue	<< 3
								);
				}

				if( paletteIndex == -1 )
				{
					// If the same tex4x4 palette does not exist
					u8	red, green, blue;								// tex4x4 palette color element

					// Set new palette address
					pColorPaletteAddress->SetValue( (nNewColor >> 1 ), block );

					// Added tex4x4 palette
					for( u8 n=0; n<nCompareColor; n++ )
					{
						fourColorPalette.GetValue(  &red, &green, &blue , n );	// Get tex4x4 palette color
						pColorPalette->SetValue( red, green, blue, nNewColor );	// Added to palette
						nNewColor++;
					}

					// Align the number of palettes to an even number.
					if( ( nCompareColor == 1 ) || ( nCompareColor == 3 ) )
					{
						pColorPalette->SetValue( 0, 0, 0, nNewColor );	// Added dummy to palette
						nNewColor++;
					}
				}
				else
				{
					// If you can find, set the existing palette address.
					pColorPaletteAddress->SetValue( (paletteIndex >> 1), block );
				}
			}


			//--------------------------------------------------------------
			//	Create color index
			//--------------------------------------------------------------
			if( fourColorPalette.GetNumber() == 2 )
			{
				ConvertLinearColorPalette_Tex4x4( &fourColorPalette, transparencyFlg );	// Create tex4x4 linearly interpolated palette
			}

			for( u32 block_y = 0; block_y<TLB_4; block_y++)
			{
				for( u32 block_x = 0; block_x<TLB_4; block_x++)
				{
					int index;
					u8 alpha;

					blockAlphaImage.GetValue( &alpha, block_x, block_y  );			// Get alpha value

					if( alpha < 0x80 )
					{
						index = 3;			// Transparent color
					}
					else
					{
						u8 red, green, blue;

						blockDirectColorImage.GetValue( &red, &green, &blue, block_x, block_y  );	// Get image color value
						index = FindNearColor( &fourColorPalette, red, green, blue, -1 );			// Search color index from palette
						if( index == -1) return false;
					}
					pColorIndexImage->SetValue( (u8)index, (x + block_x), (y + block_y)  );	// Put index value to color index
				}
			}
		}
	}

	// Change the number of palette colors from the default number to the actual number.
	pColorPalette->ChangeNumber( nNewColor );

	//--------------------------------------------------------------
	//	If the number of palette colors exceeds palette memory,
	//	compress the number of palette colors until they can be contained in the palette memory.
	//--------------------------------------------------------------
	if( nNewColor > ( TLB_PALETTEADDRESS_MAX * 2 ) )
	{
		u8	commonRed	= m_convertFactor.commonColorRange.red;
		u8	commonGreen = m_convertFactor.commonColorRange.green;
		u8	commonBlue	= m_convertFactor.commonColorRange.blue;
		u8	escapeEndlessRoopCount	= 0;	// Counter to avoid an infinite loop
		u8	compressSpeedFactor		= 1;	// Compression speed (The variable for expanding the common range. Only use if the palette count is excessive.

		while( pColorPalette->GetNumber() > ( TLB_PALETTEADDRESS_MAX * 2 ) )
		{
			// Expand the common range
			if( commonRed < ( 255 - compressSpeedFactor) )		// So you do not exceed the maximum (255)
			{
				commonRed	+= compressSpeedFactor;
			}
			if( commonGreen < ( 255 - compressSpeedFactor) )
			{
				commonGreen	+= compressSpeedFactor;
			}
			if( commonBlue < ( 255 - compressSpeedFactor) )
			{
				commonBlue	+= compressSpeedFactor;
			}

			if(!CompressColorNumber_Tex4x4( 
						pColorPalette,
						pColorPaletteAddress,
						pColorPaletteAttribute,
						commonRed << 3,
						commonGreen << 3,
						commonBlue << 3
						)) return false;

			// To avoid an infinite loop (just in case)
			escapeEndlessRoopCount++;
			if( escapeEndlessRoopCount >30 ) return false;
		}
	}

	return true;
}

} //namespace



