/****************************************************************************
  Project: NITRO-SampleTools - TXLib
  File:    TXLib_DefaultColorProcessor_Tex4x4.cpp

  Copyright 2004,2005 Nintendo Co.,Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  This is a TXLib_DefaultColorProcessor function. It is a color processor for converting to tex4x4.

  $Log: TXLib_DefaultColorProcessor_Tex4x4.cpp,v $
  Revision 1.7  2005/06/20 06:54:11  mominoki
  Changed copyright. Made revision adjustments. 

  Revision 1.1  2004/04/15		mominoki
  Corrected comment

  Revision 1.0  2004/03/03		mominoki
  Initial registration

  $NoKeywords: $
 ***************************************************************************/
#define TLB_EXPORT_

#include <cstring>
#include <cassert>
#include <TXLib_Define.h> 
#include <TXLib/Convert/TXLib_DefaultColorProcessor.h>

namespace TXLib{

/*********************************************************************************************************
	TXLib::CDefaultColorProcessor::SetPalette_Tex4x4

	Sets a palette that has more than 4 colors to a tex4x4 palette.

	In the tex4x4 palette, color order is reordered as follows.

	Palette0: 	Color with highest brightness level
	Palette1: 	Color with lowest brightness level
	Palette2: 	Color with brightness level after Palette0
	Palette3: 	Color with brightness level before Palette1

	In transparent color mode, even if there are 4 colors, only 3 will be set.
*********************************************************************************************************/
bool CDefaultColorProcessor::SetPalette_Tex4x4( 
			CColorPalette*			pNewColorPalette,			// Newly created palette
	const	CColorPalette* 			pOldColorPalette,			// Current color palette
			bool					transparencyFlg				// Transparent color mode flag
)
{
	assert( (pOldColorPalette != NULL) && (pNewColorPalette != NULL ) );
	assert( pOldColorPalette->IsAvailable() && pNewColorPalette->IsAvailable() );

	CColorPalette	palette;
	palette	= *pOldColorPalette;					// Copy current palette (to sort in descending order of brightness)

	u32 nOldColor = palette.GetNumber();

	switch( nOldColor )
	{
		u8	red, green, blue;

		case 0:			// Only transparent color
			pNewColorPalette->ChangeNumber(0);
			break;
		case 1:
			palette.GetValue( &red, &green, &blue, 0 );
			pNewColorPalette->SetValue( red, green, blue, 0 );

			pNewColorPalette->ChangeNumber(1);
			break;
		case 2:
			palette.QuickSortTopDown( true );				// Arrange color memory in descending order of brightness

			palette.GetValue( &red, &green, &blue, 0 );
			pNewColorPalette->SetValue  ( red, green, blue, 0  );

			palette.GetValue( &red, &green, &blue, nOldColor-1 );
			pNewColorPalette->SetValue  ( red, green, blue, 1  );

			pNewColorPalette->ChangeNumber(2);
			break;
		case 3:
			palette.QuickSortTopDown( true );				// Arrange color memory in descending order of brightness

			palette.GetValue( &red, &green, &blue, 0 );
			pNewColorPalette->SetValue  ( red, green, blue, 0 );

			palette.GetValue( &red, &green, &blue, nOldColor-1 );	// Pay attention to the rearrangement
			pNewColorPalette->SetValue  ( red, green, blue, 1  );

			palette.GetValue( &red, &green, &blue, nOldColor-2 );
			pNewColorPalette->SetValue  ( red, green, blue, 2  );

			pNewColorPalette->ChangeNumber(3);
			break;
		case 4:	// 4 color
			palette.QuickSortTopDown( true );				// Arrange color memory in descending order of brightness

			palette.GetValue( &red, &green, &blue, 0 );
			pNewColorPalette->SetValue  ( red, green, blue, 0 );

			palette.GetValue( &red, &green, &blue, nOldColor-1  );	// Pay attention to the rearrangement
			pNewColorPalette->SetValue  ( red, green, blue, 1 );

			palette.GetValue( &red, &green, &blue, nOldColor-3  );
			pNewColorPalette->SetValue  ( red, green, blue , 2 );

			if( !transparencyFlg )	// 4 colors if 4 color mode
			{
				palette.GetValue( &red, &green, &blue, nOldColor-2  );
				pNewColorPalette->SetValue  ( red, green, blue, 3  );

				pNewColorPalette->ChangeNumber(4);
			}
			else				// 3 colors if transparent color mode
			{
				pNewColorPalette->ChangeNumber(3);
			}
			break;
		default:
			false;
	}

	return true;
}

/*********************************************************************************************************
	TXLib::CDefaultColorProcessor::Get4ColorHiLow_Tex4x4

	Creates a tex4x4 palette using brightest, least bright, medium bright colors.
*********************************************************************************************************/
bool CDefaultColorProcessor::Get4ColorHiLow_Tex4x4( 
			CColorPalette*				pNewColorPalette,				// Newly created palette
	const	CColorPalette* 				pOldColorPalette,				// Current color palette
			bool						transparencyFlg,				// Transparent color mode flag
			bool						linearFlg						// Linear interpolation flag
)
{
	assert( (pOldColorPalette != NULL) && (pNewColorPalette != NULL ) );

	pNewColorPalette->ChangeNumber(4);	// Change the number of colors to the maximum (4)

	TXLib::CColorPalette	colorMemory;
	colorMemory = *pOldColorPalette;						// Copy current palette for processing

	colorMemory.QuickSortTopDown( true );					// Line up in descending order of brightness

	u32 hiYNumber, lowYNumber;								// Index of the colors with the hightest and lowest level of brightness
	hiYNumber = 0;
	lowYNumber = colorMemory.GetNumber() - 1;

	u8 red, green, blue;
	// Set the color with the highest level of brightness to Color0 and lowest level of brightness to Color1
	colorMemory.GetValue( &red, &green, &blue, hiYNumber );
	pNewColorPalette->SetValue( red, green, blue, 0 );

	colorMemory.GetValue( &red, &green, &blue, lowYNumber );
	pNewColorPalette->SetValue( red, green, blue, 1 );

	u32 midHiYNumber, midLowYNumber;									// Index of color with middle level of brightness

	if( linearFlg )
	{
		// Because there are only 2 colors with linear interpolation, break out of function by setting palette count to 2 colors

		pNewColorPalette->ChangeNumber(2);
	}
	else
	{
		// Without linear interpolation, set the remaining 1 color, if in transparent color mode, or 2 colors, if in 4 color mode.
		if( transparencyFlg )						// Transparent color mode
		{
			midHiYNumber = ( colorMemory.GetNumber() - 1 ) / 2 ;		// Center

			colorMemory.GetValue( &red, &green, &blue, midHiYNumber );
			pNewColorPalette->SetValue( red, green, blue, 2 );

			pNewColorPalette->ChangeNumber(3);
		}
		else
		{	
			// 4 color mode
			midHiYNumber = ( colorMemory.GetNumber() - 1 ) / 3 ;		//One-third of the highest brightness level
			midLowYNumber =(( colorMemory.GetNumber() - 1 ) * 2 ) / 3 ;	// Two-third of the highest brightness level

			colorMemory.GetValue( &red, &green, &blue, midHiYNumber );
			pNewColorPalette->SetValue( red, green, blue, 2 );

			colorMemory.GetValue( &red, &green, &blue, midLowYNumber );
			pNewColorPalette->SetValue( red, green, blue, 3 );
		}
	}

	return true;
}

/*********************************************************************************************************
	TXLib::CDefaultColorProcessor::Get4ColorRoundRobin_Tex4x4

	Compress the number of colors using the round-robin method, and create tex4x4 palette.
	Only 2 colors can be used when linear interpolation is specified.
*********************************************************************************************************/
bool
CDefaultColorProcessor::Get4ColorRoundRobin_Tex4x4( 
				CColorPalette*		pNewColorPalette,				// Newly created palette
		const	CColorPalette*		pOldColorPalette,				// Current color palette
				bool				transparencyFlg,				// Transparent color mode flag
				bool				linearFlg						// Linear interpolation flag
)
{
	assert( (pOldColorPalette != NULL) && (pNewColorPalette != NULL ) );
	assert( pOldColorPalette->IsAvailable() && pNewColorPalette->IsAvailable() );

	// Calculate the number of colors in a tex4x4 palette
	u32 nNewColor = ( transparencyFlg ) ? 3: 4;

	if( linearFlg )
	{
		nNewColor = 2;
	}

	u32 nOldColor = pOldColorPalette->GetNumber();		// Number of current palette
	pNewColorPalette->ChangeNumber(nNewColor);			// Changed number of palette

 	CColorPalette	fourPalette;							// tex4x4 palette
	fourPalette.Create( 4, 0, 0, 0 );

	u8		red, green, blue;							// Color value
	u8		paletteRed, paletteGreen, paletteBlue;
	s32		distanceWeightRed, distanceWeightGreen, distanceWeightBlue;	// Difference between weighted color values.

	u32		colorIndexMin[4];					// Index number of minimum brightness level
	u32		length, lengthMin;
	u32		lengthSum			= 0;			// Minimum value of difference in brightness levels
	u32		lengthSumMin		= 0xffffffff;	// Initial value MAX
	bool	finishFlg			= false;		// Finish flag

	// Searches for the 4 colors that have the minimum color distance (round robin).
	if( nNewColor == 4 )
	{	// 4 colors mode

		for( u32 i=0; i<( nOldColor - 3); i++ )
		{
			pOldColorPalette->GetValue( &red, &green, &blue, i );
			fourPalette.SetValue( red, green, blue,  3 );

			for( u32 j=i+1; j<(nOldColor - 2); j++ )
			{
				pOldColorPalette->GetValue( &red, &green, &blue, j );
				fourPalette.SetValue( red, green, blue,  2 );

				for( u32 k=j+1; k<(nOldColor - 1); k++ )
				{
					pOldColorPalette->GetValue( &red, &green, &blue, k );
					fourPalette.SetValue( red, green, blue,  1 );

					for( u32 l=k+1; l<nOldColor; l++ )
					{
						pOldColorPalette->GetValue( &red, &green, &blue, l );
						fourPalette.SetValue( red, green, blue,  0 );

						lengthSum = 0;
						for( u32 ii=0; ii<nOldColor; ii++ )
						{
							lengthMin = 0xffffffff;
							pOldColorPalette->GetValue( &paletteRed, &paletteGreen, &paletteBlue, ii );

							for( u32 jj=0; jj<nNewColor; jj++ )
							{
								fourPalette.GetValue( &red, &green, &blue, jj );

								distanceWeightRed	= TLB_WEIGHT_RED	* ( red		- paletteRed );
								distanceWeightGreen	= TLB_WEIGHT_GREEN	* ( green	- paletteGreen );
								distanceWeightBlue	= TLB_WEIGHT_BLUE	* ( blue	- paletteBlue );

								length = distanceWeightRed * distanceWeightRed  +
										distanceWeightGreen * distanceWeightGreen +
										distanceWeightBlue * distanceWeightBlue;

								if( length < lengthMin)
								{
									lengthMin = length;
									if( length == 0 ) break;
								}			
							}
							lengthSum += lengthMin;
						}

						if( lengthSum < lengthSumMin )			// The palette with the smallest difference becomes the tex4x4 palette
						{
							colorIndexMin[0] = l;
							colorIndexMin[1] = k;
							colorIndexMin[2] = j;
							colorIndexMin[3] = i;
							lengthSumMin = lengthSum;	// Exchange minimum value
							if( lengthSum == 0 ) finishFlg = true;
						}
						if( finishFlg )	break;
					} // for(l)
					if( finishFlg )	break;
				} // for(k)
				if( finishFlg )	break;
			}	// for(j)
			if( finishFlg )	break;
		}	// for(i)
	}
	else if( nNewColor == 3 )
	{
		// 3 colors + transparent color mode
		for( u32 j=0; j<(nOldColor - 2); j++ )
		{
			pOldColorPalette->GetValue( &red, &green, &blue, j );
			fourPalette.SetValue( red, green, blue,  2 );

			for( u32 k=j+1; k<(nOldColor - 1); k++ )
			{
				pOldColorPalette->GetValue( &red, &green, &blue, k );
				fourPalette.SetValue( red, green, blue,  1 );

				for( u32 l=k+1; l<nOldColor; l++ )
				{
					pOldColorPalette->GetValue( &red, &green, &blue, l );
					fourPalette.SetValue( red, green, blue,  0 );

					lengthSum = 0;
					for( u32 ii=0; ii<nOldColor; ii++ )
					{
						lengthMin = 0xffffffff;
						pOldColorPalette->GetValue( &paletteRed, &paletteGreen, &paletteBlue, ii );


						for( u32 jj=0; jj<nNewColor; jj++ )
						{
							fourPalette.GetValue( &red, &green, &blue, jj );

							distanceWeightRed	= TLB_WEIGHT_RED	* ( red		- paletteRed );
							distanceWeightGreen	= TLB_WEIGHT_GREEN	* ( green	- paletteGreen );
							distanceWeightBlue	= TLB_WEIGHT_BLUE	* ( blue	- paletteBlue );

							length = distanceWeightRed * distanceWeightRed  +
									distanceWeightGreen * distanceWeightGreen +
									distanceWeightBlue * distanceWeightBlue;

							if( length < lengthMin)
							{
								lengthMin = length;
								if( length == 0 ) break;
							}			
						}
						lengthSum += lengthMin;
					}

					if( lengthSum < lengthSumMin )			// The palette with the smallest difference becomes the tex4x4 palette
					{
						colorIndexMin[0] = l;
						colorIndexMin[1] = k;
						colorIndexMin[2] = j;
						lengthSumMin = lengthSum;	// Exchange minimum value
						if( lengthSum == 0 ) finishFlg = true;
					}
					if( finishFlg )	break;
				} // for(l)
				if( finishFlg )	break;
			} // for(k)
			if( finishFlg )	break;
		}	// for(j)
	}
	else if( nNewColor == 2 ) // With linear interpolation
	{
		for( u32 k=0; k<(nOldColor - 1); k++ )
		{
			pOldColorPalette->GetValue( &red, &green, &blue, k );
			fourPalette.SetValue( red, green, blue,  1 );

			for( u32 l=k+1; l<nOldColor; l++ )
			{
				pOldColorPalette->GetValue( &red, &green, &blue, l );
				fourPalette.SetValue( red, green, blue,  0 );

				lengthSum = 0;
				for( u32 ii=0; ii<nOldColor; ii++ )
				{
					lengthMin = 0xffffffff;
					pOldColorPalette->GetValue( &paletteRed, &paletteGreen, &paletteBlue, ii );
					ConvertLinearColorPalette_Tex4x4( &fourPalette, transparencyFlg );

					for( u32 jj=0; jj<fourPalette.GetNumber(); jj++ )	// 3 colors or 4 colors
					{
						fourPalette.GetValue( &red, &green, &blue, jj );

						distanceWeightRed	= TLB_WEIGHT_RED	* ( red		- paletteRed );
						distanceWeightGreen	= TLB_WEIGHT_GREEN	* ( green	- paletteGreen );
						distanceWeightBlue	= TLB_WEIGHT_BLUE	* ( blue	- paletteBlue );

						length = distanceWeightRed * distanceWeightRed  +
								distanceWeightGreen * distanceWeightGreen +
								distanceWeightBlue * distanceWeightBlue;

						if( length < lengthMin)
						{
							lengthMin = length;
							if( length == 0 ) break;
						}			
					}
					lengthSum += lengthMin;
				}

				if( lengthSum < lengthSumMin )	// The palette with the smallest difference becomes the tex4x4 palette
				{
					colorIndexMin[0] = l;
					colorIndexMin[1] = k;
					lengthSumMin = lengthSum;	// Exchange minimum value
					if( lengthSum == 0 ) finishFlg = true;
				}
				if( finishFlg )	break;
			} // for(l)
			if( finishFlg )	break;
		} // for(k)
	}

	// Rewrite palette
	for( u32 i=0; i<nNewColor; i++)
	{
		u32 index;

		index = colorIndexMin[i];
		pOldColorPalette->GetValue( &red, &green, &blue, index );
		pNewColorPalette->SetValue( red, green, blue, i );
	}

	pNewColorPalette->QuickSortTopDown( true );

	u8 redBuf, greenBuf, blueBuf;

	// Rearrange palette
	pNewColorPalette->GetValue( &red,		&green,		&blue,		( nNewColor - 1 )	);	// The lowest brightness level
	pNewColorPalette->GetValue( &redBuf,	&greenBuf,	&blueBuf,	1					);
	pNewColorPalette->SetValue( red,		green,		blue,		1					);
	pNewColorPalette->SetValue( redBuf,		greenBuf,	blueBuf,	( nNewColor - 1 )	);

	if(nNewColor == 4)
	{
		pNewColorPalette->GetValue( &red,		&green,		&blue,		2					);
		pNewColorPalette->GetValue( &redBuf,	&greenBuf,	&blueBuf,	3					);
		pNewColorPalette->SetValue( red,		green,		blue,		3					);
		pNewColorPalette->SetValue( redBuf,		greenBuf,	blueBuf,	2					);
	}

	return true;
}

/*********************************************************************************************************
	TXLib::CDefaultTextureCompressor::MakeBlockDirectColorImage_Tex4x4

	Creates a one block (4x4 Texel) direct color image from a direct color image.

	Extracts one block from the to block top left coordinate (x, y)
*********************************************************************************************************/
void	
CDefaultColorProcessor::MakeBlockDirectColorImage_Tex4x4(
			CDirectColorImage*			pBlockDirectColorImage,			// 1 block image
	const	CDirectColorImage*			pSrcDirectColorImage,			// Original image
			u16							x,								// Block starting position of original image (X Coordinate)
			u16							y								// Block starting position of original image (Y Coordinate)
)
{
	assert( ( pBlockDirectColorImage != NULL ) && ( pSrcDirectColorImage != NULL ));
	assert( pBlockDirectColorImage->IsAvailable() && pSrcDirectColorImage->IsAvailable() );

	u16 width	= pSrcDirectColorImage->GetWidth();
	u16 height	= pSrcDirectColorImage->GetHeight();

	assert( ( ( x + TLB_4 ) <= width ) && ((y + TLB_4 ) <= height) );			// Checks if the start position is proper

	for( u32 block_y=0; block_y<TLB_4; block_y++)
	{
		for( u32 block_x=0; block_x<TLB_4; block_x++)
		{
			u8 red, green, blue;

			pSrcDirectColorImage->GetValue(  &red, &green, &blue, (x + block_x), (y + block_y) );	// Gets image color values
			pBlockDirectColorImage->SetValue( red, green, blue, block_x, block_y  );				// Set to 1 block image
		}
	}
}

/*********************************************************************************************************
	TXLib::CDefaultTextureCompressor::MakeBlockAlphaImageUseTransparentColor_Tex4x4

	Creates a one block alpha image from transparent color

	Compares the transparent color to the color values in the image. If the colors are identical, it sets an alpha value to "0" (transparent).
	If no transparent color is used in the block, it sets the transparent color mode flag to false and returns it.
*********************************************************************************************************/
bool	
CDefaultColorProcessor::MakeBlockAlphaImageUseTransparentColor_Tex4x4(
			CAlphaImage*				pBlockAlphaImage,				// 1 block alpha
	const	CDirectColorImage*			pBlockDirectColorImage,			// 1 block image
			tTLBRGB						transparentColor				// Transparent color
)
{
	assert( ( pBlockAlphaImage != NULL ) && ( pBlockDirectColorImage != NULL ));
	assert( ( pBlockAlphaImage->GetWidth() == TLB_4 ) && ( pBlockAlphaImage->GetHeight() == TLB_4 ) );
	assert( ( pBlockDirectColorImage->GetWidth() == TLB_4 ) && ( pBlockDirectColorImage->GetHeight() == TLB_4 ) );

	bool transparencyFlg = true;

	u32 n = 0;
	u8	red, green, blue;

	for( u32 block_y=0; block_y<TLB_4; block_y++)
	{
		for( u32 block_x=0; block_x<TLB_4; block_x++)
		{
			pBlockDirectColorImage->GetValue( &red, &green, &blue, block_x, block_y  );		// Get color value for image

			// Create an alpha image from transparent color
			if( (transparentColor.red == red ) && ( transparentColor.green == green ) &&
					(transparentColor.blue == blue ) )
			{
				pBlockAlphaImage->SetValue( 0x00 , block_x, block_y );									// Set in 1 block of alpha
			}
			else
			{
				pBlockAlphaImage->SetValue( 0xff , block_x, block_y );									// Set in 1 block of alpha
				n++;
			}
		}
	}

	if( n == ( TLB_4 * TLB_4 ) )
	{
		transparencyFlg = false;
	}

	return transparencyFlg;
}

/*********************************************************************************************************
	TXLib::CDefaultTextureCompressor::MakeBlockAlphaImageUseAlpha_Tex4x4

	Create a one block alpha image from an alpha image.

	Compares borderValue. If the value is less than that, it sets the alpha value to "0" (transparent).
	If no transparent color is used in the block (all are larger than the borderValue), it sets the transparent color mode flag to false and returns it.
*********************************************************************************************************/
bool
CDefaultColorProcessor::MakeBlockAlphaImageUseAlpha_Tex4x4(
			CAlphaImage*				pBlockAlphaImage,				// 1 block alpha
	const	CAlphaImage*				pSrcAlphaImage,					// Original alpha
			u16							x,								// Source alpha block start point (X coordinate)
			u16							y,								// Source alpha block start point (Y coordinate)
			u8							borderValue						// Border for binarizing alpha values
)
{
	assert( ( pBlockAlphaImage != NULL ) && ( pSrcAlphaImage != NULL ));
	assert( ( pBlockAlphaImage->GetWidth() == TLB_4 ) && ( pBlockAlphaImage->GetHeight() == TLB_4 ) );

	bool transparencyFlg = true;

	u16 width	= pSrcAlphaImage->GetWidth();
	u16 height	= pSrcAlphaImage->GetHeight();

	assert( ( ( x + TLB_4 ) <= width ) && ((y + TLB_4 ) <= height) );			// Check if start position is proper

	u32 n=0;
	for( u32 block_y=0; block_y<TLB_4; block_y++)
	{
		for( u32 block_x=0; block_x<TLB_4; block_x++)
		{
			u8 alpha;

			pSrcAlphaImage->GetValue( &alpha, (x + block_x), (y + block_y)  );		// Get alpha value

			if( alpha < borderValue )
			{
				pBlockAlphaImage->SetValue( 0x00, block_x, block_y );			// Set in the one block alpha image
			}
			else
			{
				pBlockAlphaImage->SetValue( 0xff, block_x, block_y );			// Set in the one block alpha image
				n++;
			}
		}
	}

	if( n == ( TLB_4 * TLB_4 ) )
	{
		transparencyFlg = false;
	}

	return transparencyFlg;
}

/*********************************************************************************************************
	TXLib::CDefaultColorProcessor::Get4Color_Tex4x4

	Get a tex4x4 palette from one block
*********************************************************************************************************/
bool
CDefaultColorProcessor::Get4Color_Tex4x4(
			CColorPalette*				pColorPalette,					// Color palette
			bool						transparencyFlg,				// Transparent color mode flag
	const	CDirectColorImage*			pBlockDirectColorImage,			// One block image
	const	CAlphaImage*				pBlockAlphaImage,				// 1 block alpha
			tAlgorithm					algorithm						// Algorithm
) 
{
	assert( (pColorPalette != NULL ) && ( pBlockDirectColorImage != NULL ) && ( pBlockAlphaImage != NULL) );
	assert( pColorPalette->IsAvailable() && pBlockDirectColorImage->IsAvailable() );
	assert( ( pBlockDirectColorImage->GetWidth() == TLB_4 ) && ( pBlockDirectColorImage->GetHeight() == TLB_4 ) );
	assert( ( pBlockAlphaImage->GetWidth() == TLB_4 ) && ( pBlockAlphaImage->GetHeight() == TLB_4 ) );

	CColorPalette		tex4x4Palette;		// tex4x4 palette

	tex4x4Palette.Create( (TLB_4 * TLB_4), 0, 0, 0 );		// Secure one block (16 colors) of color memory.

	// Place image that was alpha extracted from a one block image into color memory. (So that the transparent color will not be included in the color reduction algorithm)
	CreateColorMemory( &tex4x4Palette, pBlockDirectColorImage, pBlockAlphaImage, 0x80 );

	tex4x4Palette.CommonLowerBit(5);						// Pseudo convert 8-bit color values to 5 bits.

	CColorPalette	tex4x4PaletteBuf = tex4x4Palette;
	tex4x4PaletteBuf.CutCommonColor();						// Cut common color

	if( tex4x4PaletteBuf.GetNumber() < 3 )
	{
		// If no more than 2 colors, place in palette unchanged.
		if(!SetPalette_Tex4x4( pColorPalette, &tex4x4PaletteBuf, transparencyFlg )) return false;
	}
	else
	{
		// If there are 2 or more colors, maximum color reduction process. Use algorithm to switch to select several processing methods.
		switch( algorithm )
		{
			case kAlgorithm_fastNotLinear:
				if( !Get4ColorHiLow_Tex4x4( pColorPalette, &tex4x4Palette, transparencyFlg, false )) return false;			// Fast. No linear interpolation
				break;
			case kAlgorithm_fastLinear:
				if( !Get4ColorHiLow_Tex4x4( pColorPalette, &tex4x4Palette, transparencyFlg, true )) return false;			// Fast. With linear interpolation
				break;
			case kAlgorithm_roundRobinNotLinear:
				if( !Get4ColorRoundRobin_Tex4x4( pColorPalette, &tex4x4Palette, transparencyFlg, false  )) return false;	// Round robin. No linear interpolation.
				break;
			case kAlgorithm_roundRobinLinear:
			default:
				if( !Get4ColorRoundRobin_Tex4x4( pColorPalette, &tex4x4Palette, transparencyFlg, true  )) return false;	// Round robin. With linear interpolation
				break;
		}
	}
	return true;
}

/*********************************************************************************************************
	TXLib::CDefaultTextureCompressor::ConvertLinearColorPalette_Tex4x4

	Create tex4x4 linearly interpolated palette from 2 colors.

	For 4 color mode
	color2 = (( 5 * color0 ) + ( 3 * color1 )) / 8
	color3 = (( 3 * color0 ) + ( 5 * color1 )) / 8

	For transparent mode
	color2 = ( color0 + color1 ) / 2
*********************************************************************************************************/
void
CDefaultColorProcessor::ConvertLinearColorPalette_Tex4x4(
			CColorPalette*				pColorPalette,					// Color palette
			bool						transparencyFlg					// Transparent color mode flag
)
{
	assert( pColorPalette );
	assert( pColorPalette->IsAvailable() );

	u8 red0, green0, blue0;
	u8 red1, green1, blue1;
	u8 red, green, blue;

	pColorPalette->GetValue( &red0, &green0, &blue0, 0  );
	pColorPalette->GetValue( &red1, &green1, &blue1, 1 );

	if( !transparencyFlg )
	{
		pColorPalette->ChangeNumber(4);

		red		= TLB_GET_TEX4X4LINEARPALETTE_COLOR2_FOUR( red0,	red1	);
		green	= TLB_GET_TEX4X4LINEARPALETTE_COLOR2_FOUR( green0,	green1	);
		blue	= TLB_GET_TEX4X4LINEARPALETTE_COLOR2_FOUR( blue0,	blue1	);

		pColorPalette->SetValue( red, green, blue, 2 );

		red		= TLB_GET_TEX4X4LINEARPALETTE_COLOR3_FOUR( red0,	red1	);
		green	= TLB_GET_TEX4X4LINEARPALETTE_COLOR3_FOUR( green0,	green1	);
		blue	= TLB_GET_TEX4X4LINEARPALETTE_COLOR3_FOUR( blue0,	blue1	);

		pColorPalette->SetValue( red, green, blue, 3 );
	}
	else
	{
		pColorPalette->ChangeNumber(3);

		red		= TLB_GET_TEX4X4LINEARPALETTE_COLOR2( red0,		red1	);
		green	= TLB_GET_TEX4X4LINEARPALETTE_COLOR2( green0,	green1	);
		blue	= TLB_GET_TEX4X4LINEARPALETTE_COLOR2( blue0,	blue1	);

		pColorPalette->SetValue( red, green, blue, 2 );
	}
}

/*********************************************************************************************************
	TXLib::CDefaultColorProcessor::CompressColorNumber_Tex4x4

	Compress palette number
	Compress palette number with specified common range value (0 - 255)
*********************************************************************************************************/
bool
CDefaultColorProcessor::CompressColorNumber_Tex4x4( 
					CColorPalette*				pColorPalette,					// Palette
					CByteData*					pColorPaletteAddress,			// Palette address
			const	CData*						pColorPaletteAttribute,			// Palette attribute
					u8							commonRed,						// Common range value Red
					u8							commonGreen,					// Common range value Green
					u8							commonBlue						// Common range value Blue
)
{
	assert( ( pColorPalette != NULL ) && ( pColorPaletteAttribute != NULL ) && ( pColorPaletteAddress != NULL ));
	assert( pColorPalette->IsAvailable() && pColorPaletteAddress->IsAvailable() && pColorPaletteAttribute->IsAvailable());

	CColorPalette	newColorPalette;										// Commonized palette
	newColorPalette.Create( pColorPalette->GetNumber() , 0, 0, 0 );

	u8	red, green, blue;
	u32 nNewColor = 0;		// Number of colors in the new palette.
	u32 nBlock = pColorPaletteAttribute->GetByteSize();

	for( u32 block=0;	block<nBlock;	block++ )
	{
		u32						paletteAddress;
		u8						paletteAttribute;
		u8						nPaletteColor;							// Number of palette colors in a one block (2 or 4)

		pColorPaletteAddress->GetValue( &paletteAddress, block  );			// Get palette address
		pColorPaletteAttribute->GetValue( &paletteAttribute, block );		// Get palette attribute

		nPaletteColor = ( paletteAttribute & 0x01 ) ? 2: 4;			// From with linear interpolation, or without linear interpolation, seek the number of palette colors in one palette.

		// Search to see if there is already a palette that is the same as the newly created one.
		bool	existFlg = false;								// Exist? flag. (false: do not exist, true: exist)
		u8		paletteRed, paletteGreen, paletteBlue;			// New palette color values
		u8		colorRemainder;									// Color difference

		for( u32 j=0; j<nNewColor; j += 2)
		{
			u32 existCount = 0;

			// Compare the palette colors of all newly created palettes, to 2 or 4-color.
			for( u32 i=0; i<nPaletteColor; i++ )
			{
				pColorPalette->GetValue( &red, &green, &blue , ( paletteAddress * 2 + i ) );	// Get palette color
				newColorPalette.GetValue( &paletteRed, &paletteGreen, &paletteBlue, (j + i) );

				// Possible to make same as palette color?
				colorRemainder = ( blue > paletteBlue ) ? (blue - paletteBlue): (paletteBlue - blue);	// Seek color difference

				if( colorRemainder > commonBlue ) break;		// If even one color different, another palette.

				colorRemainder = ( green > paletteGreen ) ? (green - paletteGreen): (paletteGreen - green);

				if( colorRemainder > commonGreen ) break;		// If even one color different, another palette.

				colorRemainder = ( red > paletteRed ) ? (red - paletteRed): (paletteRed - red);

				if( colorRemainder > commonRed ) break;			// If even one color different, another palette.

				existCount++;
			} // for(i)

			if( existCount == nPaletteColor )	// Exists in new palette
			{
				existFlg = true;
				pColorPaletteAddress->SetValue( ( j / 2 ), block  );	// Change palette address
				break;
			}
		} // for(j)


		// If not in the new palette, add the color value to the palette.
		if( !existFlg )
		{
			pColorPaletteAddress->SetValue( ( nNewColor / 2 ), block  );	// Set palette address

			for( u32 i=0; i<nPaletteColor; i++ )
			{
				pColorPalette->GetValue( &red, &green, &blue , ( paletteAddress * 2 + i ) );	// Get palette color
				newColorPalette.  SetValue( red,	green,	 blue, nNewColor   );
				nNewColor++;
			}
		}
	}

	newColorPalette.ChangeNumber(nNewColor);
	*pColorPalette = newColorPalette;

	return true;
}

} //namespace