/****************************************************************************
  Project: NITRO-SampleTools - TXLib
  File:    TXLibUtility_Tga.h

  Copyright 2004,2005 Nintendo Co.,Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  This is a class for inputting TGA data into CTexture, and for outputting CTexture data as TGA data.

  $Log: TXLibUtility_Tga.h,v $
  Revision 1.9  2005/06/20 06:54:11  mominoki
  Updated copyright. Made revision adjustments. 

  Revision 1.3  2005/02/21		mominoki
  Changed color palettes to support RGBA, in addition to RGB. 
  Updated Read() and ReadColorPalette().

  Revision 1.1  2004/04/15		mominoki
  Corrected commnet

  Revision 1.0  2004/03/03		mominoki
  Initial registration

  $NoKeywords: $
 ***************************************************************************/
#ifndef TXLIBUTILITY_TGA_H_
#define TXLIBUTILITY_TGA_H_

#include <TXLib.h>

namespace TXLibUtility
{
	// Class
	class CTga
	{
	public:
		// Structure
		typedef struct tFileHeader
		{ // Tga header
			u8 		idFieldByteSize;				// Byte size of ID field
			u8 		paletteFlg;						// Flag determining whether or not there is a palette format (1: YES, 2: NO)
			u8 		imageType;						// Image format (defined with #define)
			u16 	paletteEntry;					// First palette
			u16 	paletteColorNumber;				// Number of palette colors
			u8		paletteBitCount;				// Number of bits per color in palette
			u16 	imagePositionX;					// Image X origin
			u16 	imagePositionY;					// Image Y origin
			u16 	width;							// Image width
			u16 	height;							// Image height
			u8 		bitCount;						// Number of bits per pixel
			u8 		imageBit;						// Image descriptor bit
		}tFileHeader;	// TGA TGAHEADER Structure
						// imageBit
						//	7-6		5-4				3-0
						//	Prohibited		dataLineType	Attribute bit

		// Definition
		enum	tImageType{
			kImageType_not,							// No image
			kImageType_palette,						// Color index, no RLE compression
			kImageType_direct,						// Direct color, no RLE compression
			kImageType_gray,						// Black and white, no RLE compression
			kImageType_paletteRle	= 9,			// Color index, RLE compression
			kImageType_directRle,					// Direct color, RLE compression
			kImageType_grayRle,						// Black and white, RLE compression
		}; //  Input/Output data format

		enum	tDataLineType{
			kDataLineType_leftDown	= 0x00,			// Pixel: Left -> Right, Row: Bottom -> Top
			kDataLineType_rightDown = 0x10,			// Pixel: Right -> Left, Row: Bottom -> Top
			kDataLineType_leftUp 	= 0x20,			// Pixel: Left -> Right, Row: Top -> Bottom
			kDataLineType_rightUp 	= 0x30,			// Pixel: Right -> Left, Row: Top -> Bottom
		};	// Data list

		enum	tDefine{
			kDefine_headerByteSize	= 0x12,	 		// Header size
			kDefine_footerByteSize	= 26			// Footer size
		};	// Definition

	private:
		// Function
		CTga( const CTga&		rTga){};		// Copy constructor

		void UnpackRLEData(
					u8*						pUnpackMemory,					// Extracted data pointer
			const	u8*						pMemory,						// Original data
					u16						bitCount,						// Bit depth
					u16						width,							// Image size width
					u16						height							// Image size height
		) const; // Extracted RLE compressed data


		bool ReadTGAHEADER(
					tFileHeader*				pHeader,						// TGA header pointer
			const	u8*							pMemory							// Input data pointer
		); // tFileHeader input


		bool WriteTGAHEADER(
					u8*							pMemory,						// Output data pointer
					tImageType				type,							// TGA format
					u16							bitCount,						// Bit depth
					u16							width,							// Width
					u16							height,							// Height			
					u32							nColorUsed,						// Actural number of palette colors
					u8							dataLineType					// Image data list format
		) const;// Write TGA header


		void WriteFooter(
					u8*							pMemory,					// Output data pointer
					u32							extentFieldPos,				// Position of expanded region
					u32							developerFieldPos			// Position of developer directory file
		) const;// Write footer


		void ReadDirectColorImageAlpha(
					TXLib::CDirectColorImage*	pDirectColorImage,				// Direct color image pointer
					TXLib::CAlphaImage*			pAlphaImage,					// Alpha image pointer
			const	u8*							pMemory,						// Input data pointer
					u16							bitCount,						// Bit depth
					u16							width,							// Width
					u16							height,							// Height
					u8							dataLineType					// Image data list format
		); // Input memory data to DirectColorImage, Alpha


		bool WriteDirectColorImageAlpha(
					u8*							pMemory,						// Output data pointer
			const	TXLib::CDirectColorImage*	pDirectColorImage,				// Direct color image pointer
			const	TXLib::CAlphaImage*			pAlphaImage,					// Alpha image pointer
					u16							bitCount,						// Bit depth
					u8							dataLineType					// Image data list format
		) const; // Output DirectColorImage, Alpha to memory


		void ReadColorPalette(
					TXLib::CColorPalette*		pColorPalette,					// Color palette pointer
			const	u8*							pMemory,						// Input data pointer
					u32							nColorUsed,						// Actual number of palette colors
					u8							paletteBitCount					// No. of bits per palette color
		); // Input memory data to ColorPalette


		void WriteColorPalette(
					u8*							pMemory,						// Output data pointer
			const	TXLib::CColorPalette*		pColorPalette					// Color palette pointer
		) const; // Output ColorPalette to memory


		bool ReadColorIndexImage(
					TXLib::CColorIndexImage*	pColorIndexImage,				// Color index image
			const	u8*							pMemory,						// Input data pointer
					u16							width,							// Width
					u16							height,							// Height
					u8							dataLineType					// Image data list format
		);// Input the memory data to ColorIndexImage

		bool WriteColorIndexImage(
					u8*							pMemory,						// Output data pointer 
			const	TXLib::CColorIndexImage*	pColorIndexImage,				// Color index image
					u16							bitCount,						// Bit depth
					u8							dataLineType				// Image data line format 
		) const; // Output the ColorIndexImage to memory

		void ReadAlphaPalette(
					TXLib::CAlphaImage*			pAlphaImage,					// Alpha image pointer 
			const	u8*							pMemory,						// Input data pointer
					TXLib::CColorIndexImage*	pColorIndexImage				// Color index image 
		); // Input memory data to AlphaImage
	public:
		CTga(){}																// Constructor
		virtual ~CTga(){}														// Destructor

		bool Read(
					TXLib::CTexture*		pTexture,						// Output Texture pointer
			const	u8*						pMemory							// Input data pointer
		); // Load TGA data from memory


		bool Write(
					u8*						pMemory,						// Output data pointer
					u32*					pByteSize,						// Number of bytes output
					TXLib::CTexture*		pTexture,						// CTexture pointer
					u8						dataLineType					// Image data list format
		) const; // Write TGA data to memory


		TXLib::CTexture::tType GetConvertType(
					TXLib::CTexture*		pTexture							// CTexture pointer
		) const; // Get conversion format


		bool  GetDataByteSizeMax(
					u32*					pMemorySize,					// Memory size
					TXLib::CTexture*		pTexture						// CTexture pointer
		) const; //  Get lergest file size necessary for saving TGA files.

	};
};
#endif