/****************************************************************************
  Project: NITRO-SampleTools - TXLib
  File:    TXLibUtility.cpp

  Copyright 2004,2005 Nintendo Co.,Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  This group of functions has been prepared in order to make TXLib more easily accesible from Tools.
  This group of functions is a sample of how to access TXLib.

  $Log: TXLibUtility.cpp,v $
  Revision 1.8  2005/06/20 06:54:11  mominoki
  Revised copyright. Made revision adjustments. 

  Revision 1.1  2004/04/15		mominoki
  Fixed comments.

  Revision 1.0  2004/03/03		mominoki
  Entered initial version.

  $NoKeywords: $
 ***************************************************************************/
#include "stdafx.h"
#include <cstring>
#include <cstdio>
#include <cstdlib>
#include "../include/TXLibUtility.h"

namespace std{ using :: memset; }

namespace TXLibUtility{

/*********************************************************************************************************
	TXLibUtility::GetFileType()	

	Get file format from file extension
*********************************************************************************************************/
static bool
GetFileType(
			u8*							pFileType,						// File format pointer
	const 	char*						pPathName						// File path name
)
{
	assert( (pPathName != NULL ) && ( pFileType != NULL ) );

	char extName[ 10 ]; // Extension

	_splitpath( pPathName, NULL, NULL, NULL, extName);	// Take out extension
	_strlwr( extName );									// Convert extention to lowercase letters

	if( strrchr( extName, 'bmp' ) )
	{
		*pFileType = kImageType_bmp;
	}
	else if( strrchr( extName, 'tga' ) )
	{
		*pFileType = kImageType_tga;
	}
	else
	{
		return false;
	}

	return true;
}

/********************************************************************************************************
	TXLibUtility::GetFileSize()	

	Get file size
*********************************************************************************************************/
static bool
GetFileSize( 
			u32*				pByteSize,						// File size pointer
	const	char*				pPathName						// File path name
)
{
	assert( pPathName != NULL );

	// Open file
	FILE *pFile;
	if( ( pFile = fopen( pPathName, "rb" ) ) == NULL ) return false;

	if( fseek( pFile, 0, SEEK_END ) == -1) return false; // Move to end of file
	if( (*pByteSize = ftell( pFile )) == -1) return false;

	fclose( pFile ); // Close file

	return true;
}

/********************************************************************************************************
	TXLibUtility::ChangeExtName()

	Change extension of path name
*********************************************************************************************************/
bool
ChangeExtName( 
			char*				pPathName,							// Path name of change target
	const	char*				extName								// Extension to change
)
{
	assert( ( pPathName != NULL ) && ( extName != NULL ));

	char* pChar = strrchr( pPathName, '.' ); // backwards"."(extension) search

	if( pChar )
	{
		strcpy( pChar, extName );		// Change extension
	}
	else
	{
		strcat( pPathName, extName);	// Add extension
	}

	return true;
}

/********************************************************************************************************
	TXLibUtility::Load()	

	Load image file
*********************************************************************************************************/
bool
Load( 
			TXLib::CTexture*		pTexture,						// CTexture pointer
	const	char*					pPathName						// Path name of file to load
)
{
	assert( ( pPathName != NULL ) && ( pTexture != NULL ));

	bool result = false; 			// Execution result

	u8 		inputDataType;				// File format
	u32 	fileSize;					// File size
	u8*		pInputData;					// Input data

	pTexture->Clear();					// Clear CTexture

	if( !GetFileSize( &fileSize, pPathName )) return false;						// Get file size of input file

	pInputData = new u8[ fileSize ];											// Memory reservation of input data

	// Input file -> Input data memory loading
	FILE *pFile;
	if( ( pFile = fopen( pPathName, "rb" ) ) == NULL ) goto error;
	fread( pInputData, 1, fileSize, pFile );
	fclose( pFile );

	if( !GetFileType( &inputDataType, pPathName ) ) goto error;				// Get input data format from input file extension

	// Load input data, create CTexture derivation class, and copy CTexture
	if( inputDataType == kImageType_bmp )
	{
		TXLibUtility::CBmp bmp;
		if( !bmp.Read( pTexture, pInputData )) goto error;

	}
	else if( inputDataType == kImageType_tga )
	{
		TXLibUtility::CTga tga;
		if( !tga.Read( pTexture, pInputData )) goto error;
	}
	else
	{
		goto error;
	}

	result = true;
error:
	delete[] pInputData;
	pInputData = NULL;

	return result;
}

/********************************************************************************************************
	TXLibUtility::Save()	

	Save image file
*********************************************************************************************************/
bool
Save(
	const 	char*					pPathName,						// Path name of save file
			TXLib::CTexture*		pTexture						// CTexture pointer
)
{
	assert( ( pPathName != NULL ) && ( pTexture != NULL ));

	bool result = false; 			// Execution result

	// Reserve memory necessary for output
	u8*		pData;
	u32		dataByteSize;
	u8 		outputDataType;				// File format

	if( !GetFileType( &outputDataType, pPathName ) ) return false;		// Get output data format from file extension

	// Load input data, create CTexture derivation class, and copy CTexture
	if( outputDataType == kImageType_bmp )
	{
		CBmp bmp;
		if( !bmp.GetDataByteSizeMax( &dataByteSize, pTexture )) return false;	// Get memory size

		pData = new u8[ dataByteSize];
		std::memset( pData, 0, dataByteSize );

		if( !bmp.Write( pData, &dataByteSize, pTexture )) goto error;		// Write to memory
	}
	else if( outputDataType == kImageType_tga )
	{
		CTga tga;
		if( !tga.GetDataByteSizeMax( &dataByteSize, pTexture )) return false;	// Get memory size

		pData = new u8[ dataByteSize];
		std::memset( pData, 0, dataByteSize );

		if( !tga.Write( pData, &dataByteSize, pTexture, CTga::kDataLineType_leftDown )) goto error;		// Write to memory
	}
	else
	{
		goto error;
	}

	// Save file
	FILE *pFile;
	if( ( pFile = fopen( pPathName, "wb" ) ) == NULL ) goto error;
	fwrite( pData, 1, dataByteSize, pFile );
	fclose( pFile );

	result = true;
error:
	// Clear memory
	if( pData != NULL )
	{
		delete[] pData;
		pData = NULL;
	}

	return result;
}

} // namespace

