/****************************************************************************
  Project: NITRO-SampleTools - TXLib
  File:    TXLibUtility_Tga.cpp

  Copyright 2004,2005 Nintendo Co.,Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  This is a class for inputting TGA data to a CTexture, and outputting CTexture data to TGA data.

  $Log: TXLibUtility_Tga.cpp,v $
  Revision 1.9  2005/06/20 06:54:11  mominoki
  Revised copyright. Made revision adjustments. 

  Revision 1.3  2005/02/21		mominoki
  Changed color palettes to support RGBA, in addition to RGB.
  Updated Read(), ReadColorPalette().

  Revision 1.2  2004/06/25		mominoki
  Created workaround for problem of photoshop not displaying the alpha channel
  by adding 0x08 to imagebit during write.

  Revision 1.1  2004/04/15		mominoki
  Fixed comments.

  Revision 1.0  2004/03/03		mominoki
  Entered initial version.

  $NoKeywords: $
 ***************************************************************************/
#include "stdafx.h"
#include <cstring>
#include "../include/TXLibUtility_Tga.h"


namespace std{ using :: memset; }
namespace std{ using :: memcpy; }

namespace TXLibUtility{

/*********************************************************************************************************
		private
*********************************************************************************************************/
/********************************************************************************************************
	TXLibUtility::CTga::UnpackRLEData()

	Unpack RLE compression data
*********************************************************************************************************/
void
CTga::UnpackRLEData(
			u8*						pUnpackMemory,					// Unpacked data pointer
	const	u8*						pMemory,						// Original data
			u16						bitCount,						// Bit depth
			u16						width,							// Image size   Width
			u16						height							// Image size   Height
) const
{
	assert( (pUnpackMemory != NULL) && (pMemory != NULL));
	assert( ( bitCount == 8 ) || ( bitCount == 16 ) || ( bitCount == 24 ) || ( bitCount == 32 ) );

	u32		dataNumberMax = width * height;	// Largest data number
	u32		dataNumber = 0;					// Loaded data number
	u32		unpackDataOffset = 0;			// Unpack data offset
	u32		dataOffset = 0;					// Data offset
	u8		data;
	u16		lineDataNumber;					// Data number on one line	
	bool	literalFlg;						// Literal flag

	u8 byte = bitCount >> 3;	// Byte size
	while( dataNumber < dataNumberMax )
	{
		data			= *( pMemory + dataOffset );
        literalFlg		= ( data & 0x80 ) ? false : true;	//  Literal determination
		lineDataNumber	= (u16)( (data & 0x7f) + 1);		// When the loaded data number is 0, the data number is 1, so increase by one.
		dataOffset++;

		if( literalFlg )
		{
			// Literal
			for( u32 i=0; i<lineDataNumber; i++ )
			{
				for( u32 j=0; j<byte; j++ )
				{
					*( pUnpackMemory + unpackDataOffset ) = *( pMemory + dataOffset );
					dataOffset++;
					unpackDataOffset++;
				}
				dataNumber++;
			}
		}
		else
		{
			// Repeat
			u8  data[4];
			u32 i;

			for( i=0; i<byte; i++ )
			{
				data[i] = *( pMemory + dataOffset );
				dataOffset++;
			}

			for( i=0; i<lineDataNumber; i++ )
			{
				for( u32 j=0; j<byte; j++ )
				{
					*( pUnpackMemory + unpackDataOffset     ) = data[j];
					unpackDataOffset++;
				}
				dataNumber++;
			}
		}

	}
}

/********************************************************************************************************
	TXLibUtility::CTga::ReadTGAHEADER()

	Load tFileHeader from memory.

		The loadable TGA conditions are as follows:

		The starting position of the image is at the origin.

		The grouping of imageType and bitCount is as follows:
			imageType				bitCount 
			kImageType_palette		8 
			kImageType_direct		16 
			kImageType_direct		24 
			kImageType_direct		32 

		The image data line is
			Pixel: Left to right, Row: Bottom to top
			Pixel: Left to right, Row: Top to bottom
*********************************************************************************************************/
bool
CTga::ReadTGAHEADER(
			tFileHeader*				pHeader,						// TGA header pointer
	const	u8*							pMemory							//Input data pointer
)
{
	assert( ( pMemory != NULL ) && ( pHeader != NULL ) );

	u32 dataOffset = 0;

	pHeader->idFieldByteSize	= *( pMemory + dataOffset );		// ID field
	dataOffset += 1;

	pHeader->paletteFlg			= *( pMemory + dataOffset );		// Palette flag
	dataOffset += 1;

	pHeader->imageType			= *( pMemory + dataOffset );		// Format
	dataOffset += 1;

	pHeader->paletteEntry		= *(pMemory + dataOffset) | (*(pMemory + dataOffset + 1) << 8);
	dataOffset += 2;

	pHeader->paletteColorNumber	= *(pMemory + dataOffset) | (*(pMemory + dataOffset + 1) << 8); // Number of palette colors
	dataOffset += 2;

	pHeader->paletteBitCount	= *( pMemory + dataOffset );		// Palette 1-color bit depth
	dataOffset += 1;

	pHeader->imagePositionX		= *(pMemory + dataOffset) | (*(pMemory + dataOffset + 1) << 8); // X origin
	dataOffset += 2;

	pHeader->imagePositionY		= *(pMemory + dataOffset) | (*(pMemory + dataOffset + 1) << 8); // Y origin
	dataOffset += 2;

	pHeader->width				= *(pMemory + dataOffset) | (*(pMemory + dataOffset + 1) << 8); // Width
	dataOffset += 2;

	pHeader->height				= *(pMemory + dataOffset) | (*(pMemory + dataOffset + 1) << 8); // Height
	dataOffset += 2;

	pHeader->bitCount			= *( pMemory + dataOffset );		// Image 1 texel bit depth
	dataOffset += 1;

	pHeader->imageBit			 = *( pMemory + dataOffset );		// Image descriptor bit
	dataOffset += 1;

	if( ( pHeader->imagePositionX != 0 ) && ( pHeader->imagePositionY != 0 ) ) return false;		// Origin is "0"

	if( ( pHeader->imageType != kImageType_palette ) && ( pHeader->imageType != kImageType_direct ) &&
		( pHeader->imageType != kImageType_gray ) && 
		( pHeader->imageType != kImageType_paletteRle ) && ( pHeader->imageType != kImageType_directRle ) &&
		( pHeader->imageType != kImageType_grayRle ) ) return false;					// Format

	if( ( pHeader->bitCount != 8 ) && ( pHeader->bitCount != 16 ) &&
		( pHeader->bitCount != 24 ) && ( pHeader->bitCount != 32 ) ) return false;		// Bit depth

	return true;
}

/********************************************************************************************************
	TXLibUtility::CTga::WriteTGAHEADER()

	Write TGA header
*********************************************************************************************************/
bool
CTga::WriteTGAHEADER(
			u8*							pMemory,						// Output data pointer
			tImageType					type,							// TGA format
			u16							bitCount,						// Bit depth
			u16							width,							// Width
			u16							height,							// Height			
			u32							nColorUsed,						// Actual number of palette colors
			u8							dataLineType					// Image data line format
) const
{
	assert( pMemory != NULL );

	tFileHeader header;


	header.imagePositionX		= 0;						// Origin X
	header.imagePositionY		= 0;						// Origin Y
	header.imageType			= type;						// TGA format
	header.bitCount				= (u8)bitCount;				// Bit depth
	header.width				= width;					// Width
	header.height				= height;					// Height
	header.imageBit				= dataLineType;				// Image data line format
	header.idFieldByteSize		= 0;						// ID field size
	header.paletteEntry			= 0;

	switch( type )
	{
		case kImageType_palette:
			if( bitCount == 8 )
			{
				header.paletteBitCount	= 24;
			}else{
				return false;
			}
			header.paletteFlg			= 1;
			header.paletteColorNumber	= nColorUsed;
			break;
		case kImageType_gray:
		case kImageType_direct:
			if( ( bitCount == 8 ) || ( bitCount == 16 ) || ( bitCount == 24 ) || ( bitCount == 32 ) )
			{
				header.paletteFlg			= 0;
				header.paletteBitCount		= 0;
				header.paletteColorNumber	= 0;
			}
			else
			{
				return false;
			}
			break;
		default:
			return false;
	}


	u32 dataOffset=0;

	*( pMemory + dataOffset )			= header.idFieldByteSize;
	dataOffset += 1;

	 *( pMemory + dataOffset )			= header.paletteFlg;
	dataOffset += 1;

	*( pMemory + dataOffset )			= header.imageType;
	dataOffset += 1;

	*(pMemory + dataOffset    )			= (u8)(header.paletteEntry & 0xff);
	*(pMemory + dataOffset + 1)			= (u8)(header.paletteEntry >> 8);
	dataOffset += 2;

	*(pMemory + dataOffset    )			= (u8)(header.paletteColorNumber & 0xff);
	*(pMemory + dataOffset + 1)			= (u8)(header.paletteColorNumber >> 8);
	dataOffset += 2;

	*( pMemory + dataOffset )			= header.paletteBitCount;
	dataOffset += 1;

	*(pMemory + dataOffset    )			= (u8)(header.imagePositionX & 0xff);
	*(pMemory + dataOffset + 1)			= (u8)(header.imagePositionX >> 8);
	dataOffset += 2;

	*(pMemory + dataOffset    )			= (u8)(header.imagePositionY & 0xff);
	*(pMemory + dataOffset + 1)			= (u8)(header.imagePositionY >> 8);
	dataOffset += 2;

	*(u16*)( pMemory + dataOffset )		= header.width;
	dataOffset += 2;

	*(u16*)( pMemory + dataOffset )		= header.height;
	dataOffset += 2;

	*( pMemory + dataOffset )			= header.bitCount;
	dataOffset += 1;

	*( pMemory + dataOffset )			= header.imageBit | 0x08;
	dataOffset += 1;

	return true;
}

/********************************************************************************************************
	TXLibUtility::CTga::WriteFooter()

	 Output footer to memory
*********************************************************************************************************/
void
CTga::WriteFooter(
			u8*							pMemory,					// Output data pointer
			u32							extentFieldPos,				// Position of expanded region
			u32							developerFieldPos			// Position of developer directory file
) const
{
	assert( pMemory != NULL );

	*pMemory			= (u8)(extentFieldPos & 0xff);
	*(pMemory + 1 )		= (u8)((extentFieldPos >> 8 ) & 0xff );
	*(pMemory + 2 )		= (u8)((extentFieldPos >> 16) & 0xff );
	*(pMemory + 3 )		= (u8)((extentFieldPos >> 24) & 0xff );

	*(pMemory + 4 )		= (u8)(developerFieldPos & 0xff);
	*(pMemory + 5 )		= (u8)((developerFieldPos >> 8 ) & 0xff );
	*(pMemory + 6 )		= (u8)((developerFieldPos >> 16) & 0xff );
	*(pMemory + 7 )		= (u8)((developerFieldPos >> 24) & 0xff );

	char strTga[18] = "TRUEVISION-TARGA";
	memcpy( pMemory + 8, strTga, 17 );
	*( pMemory + 25 )	= 0;
}

/********************************************************************************************************
	TXLibUtility::CTga::ReadDirectColorImageAlpha()

	Input memory data to DirectColorImage, Alpha

		When the image data line is Pixel: left to right, Row: Bottom to top, rearrange to Pixel: left to right, Row: top to bottom.

		Rearrange in direct order and set so that the data line agrees with CTga.

		Use as default value (0xff) without loading an alpha value when 16-bit or 24-bit.

		When 16-bit, convert to 8-bit by left-shifting each color value by 3 bits and copying the upper 3-bit value to the empty lower 3-bit value.
*********************************************************************************************************/
void
CTga::ReadDirectColorImageAlpha(
			TXLib::CDirectColorImage*	pDirectColorImage,				// Direct color image pointer
			TXLib::CAlphaImage*			pAlphaImage,					// Alpha image pointer
	const	u8*							pMemory,						// Input data pointer
			u16							bitCount,						// Bit depth
			u16							width,							// Width
			u16							height,							// Height
			u8							dataLineType					// Image data line format
)
{
	assert(( pDirectColorImage != NULL ) && ( pAlphaImage != NULL ) && ( pMemory != NULL ));

	pDirectColorImage->Create( width, height, 0, 0, 0 );
	pAlphaImage->Create( width, height, 0xff );

	u32 dataOffset;
	u8	redBuf, greenBuf, blueBuf;
	u16 data_16bit;

	for( u16 y=0; y<height; y++)
	{
		u16 yUp = y;
		if( ( dataLineType == kDataLineType_leftDown ) || (dataLineType == kDataLineType_rightDown) )
		{
			yUp = height - 1 - y;	// // Invert Y coordinate
		}

		for( u16 x=0; x<width; x++ )
		{
			u16 xRight = x;
			if( ( dataLineType == kDataLineType_rightDown ) || (dataLineType == kDataLineType_rightUp) )
			{
				xRight = width - 1 - x;	// // Invert X coordinate
			}

			switch( bitCount )
			{
				case 8: // Gray
					dataOffset = (yUp * width) + xRight;

					pDirectColorImage->SetValue(
							*( pMemory + dataOffset ),	// Red
							*( pMemory + dataOffset ),	// Green
							*( pMemory + dataOffset ),	 // Blue
							x, y);						// (x,y)
					break;

				case 16: // direct5551
					dataOffset = ( ( yUp * width) + xRight) * 2;
					data_16bit = (u16)( *(u16*)( pMemory + dataOffset ) );

					redBuf		= TLB_GET_L2_DATA_1555( data_16bit );
					greenBuf	= TLB_GET_L3_DATA_1555( data_16bit );
					blueBuf		= TLB_GET_L4_DATA_1555( data_16bit );

					pDirectColorImage->SetValue(
												//  Upper 5-bit				Lower 3-bit
												(redBuf << 3)	| ( (redBuf >> 2)	& 0x07 ),		// Red
												(greenBuf << 3) | ( (greenBuf >> 2) & 0x07 ),		// Green
												(blueBuf<< 3)	| ( (blueBuf >> 2)	& 0x07 ),		// Blue
												x, y);												// (x,y)


//					alphaBuf = GET_L1_DATA_1555( data_16bit );	// Alpha
//					alpha = ( alphaBuf > 0 ) ? 0xff : 0x00;		// Do not get, since application is not supported.
//					pAlphaImage->SetValue( alpha, x, y );		// Do not set to the same.
					break;
				case 24: // direct
					dataOffset = ( (yUp * width) + xRight) * 3;

					pDirectColorImage->SetValue(
												*( pMemory + dataOffset + 2 ),	// Red
												*( pMemory + dataOffset + 1 ),	// Green
												*( pMemory + dataOffset     ),  // Blue
												x, y);							// (x,y)

					break;
				case 32: // direct
					dataOffset = ( (yUp * width) + xRight) * 4;

					pDirectColorImage->SetValue(
												*( pMemory + dataOffset + 2 ),	// Red
												*( pMemory + dataOffset + 1 ),	// Green
												*( pMemory + dataOffset     ),	// Blue
												x, y);							// (x,y)

					pAlphaImage->SetValue( *( pMemory + dataOffset + 3 ), x, y ); // Alpha
					break;
				default:
					assert(false);
			}
		}
	}
}

/********************************************************************************************************
	TXLibUtility::CTga::WriteDirectColorImageAlpha()

	 Output DirectColorImage, Alpha to memory

	Rearrange data and set
*********************************************************************************************************/
bool
CTga::WriteDirectColorImageAlpha(
			u8*							pMemory,					// Output data pointer
	const	TXLib::CDirectColorImage*	pDirectColorImage,			// Direct color image pointer
	const	TXLib::CAlphaImage*			pAlphaImage,				// Alpha image pointer
			u16							bitCount,					// Bit depth
			u8							dataLineType				// Image data line format
) const
{
	assert(  ( pMemory != NULL ) && ( pDirectColorImage != NULL ));
	assert(	pDirectColorImage->IsAvailable());

	u16 width	= pDirectColorImage->GetWidth();
	u16 height	= pDirectColorImage->GetHeight();

	for( u16 y=0; y<height; y++)
	{
		u16 yUp = y;
		if( ( dataLineType == kDataLineType_leftDown ) || (dataLineType == kDataLineType_rightDown) )
		{
			yUp = height - 1 - y;	// // Invert Y coordinate
		}

		for( u16 x=0; x<width; x++ )
		{
			u16 xRight = x;
			if( ( dataLineType == kDataLineType_rightDown ) || (dataLineType == kDataLineType_rightUp) )
			{
				xRight = width - 1 - x;	// // Invert X coordinate
			}

			u8	red, green, blue, alpha; // Color value and alpha value
			u32 dataOffset;
			u16	data_16bit;

			pDirectColorImage->GetValue( &red, &green, &blue, x, y  );	// Get image color value

			if( pAlphaImage->IsAvailable() )
			{
				pAlphaImage->GetValue(  &alpha, x, y );						// Get alpha value
			}
			else
			{
				alpha = 0xff;
			}

			// Set color value and alpha value
			switch( bitCount )
			{
				case 16:
					red >>= 3;
					green >>= 3;
					blue >>= 3;
					alpha = (alpha > 0 ) ? 1: 0; // Alpha 8-bit to 1-bit conversion (if an alpha exists, set to "1")

					dataOffset = (( yUp * width) + xRight) * 2;
					data_16bit = TLB_DATA_1555( alpha, red, green, blue);
					*(u16*)( pMemory + dataOffset ) = data_16bit;
					break;
				case 24:
					dataOffset = (( yUp * width) + xRight) * 3;

					*( pMemory + dataOffset     ) = blue; 
					*( pMemory + dataOffset + 1 ) = green;
					*( pMemory + dataOffset + 2 ) = red;
					break;
				case 32:
					dataOffset = (( yUp * width) + xRight) * 4;

					*( pMemory + dataOffset     ) = blue; 
					*( pMemory + dataOffset + 1 ) = green;
					*( pMemory + dataOffset + 2 ) = red;
					*( pMemory + dataOffset + 3 ) = alpha;
					break;
				default:
					return false;
			}
		}
	}

	return true;
}

/********************************************************************************************************
	TXLibUtility::CTga::ReadColorPalette()

	Output memory data to ColorPalette
*********************************************************************************************************/
void
CTga::ReadColorPalette(
			TXLib::CColorPalette*	pColorPalette,					// Color palette pointer
	const	u8*						pMemory,						// Input data pointer
			u32						nColorUsed						// Actual number of palette colors
			u8						paletteBitCount					// Number of bit per palette
) 
{
	assert((pColorPalette != NULL ) && ( pMemory != NULL ));

	pColorPalette->Create( nColorUsed, 0, 0, 0 );

	if(paletteBitCount == 24)
	{
	for( u32 i=0; i<nColorUsed; i++)
	{
		u8 blue =	*( pMemory + ( i * 3 )     );			// Get
		u8 green =	*( pMemory + ( i * 3 ) + 1 );
		u8 red =	*( pMemory + ( i * 3 ) + 2 );

			pColorPalette->SetValue( red, green, blue, i  );	// Set
		}
	}
	else if(paletteBitCount == 32)
	{
		for( u32 i=0; i<nColorUsed; i++)
		{
			u8 blue =	*( pMemory + ( i * 4 )     );			// Get
			u8 green =	*( pMemory + ( i * 4 ) + 1 );
			u8 red =	*( pMemory + ( i * 4 ) + 2 );

			pColorPalette->SetValue( red, green, blue, i  );	// Set
		}
	}
}

/********************************************************************************************************
	TXLibUtility::CTga::WriteColorPalette()

	Output ColorPalette to memory
*********************************************************************************************************/
void
CTga::WriteColorPalette(
			u8*						pMemory,						// Output data pointer
	const	TXLib::CColorPalette*	pColorPalette					// Color palette pointer
) const
{
	assert(( pMemory != NULL ) && ( pColorPalette != NULL ));
	assert( pColorPalette->IsAvailable());

	for( u32 i=0; i<pColorPalette->GetNumber(); i++)
	{
		u8 red, green, blue;

		pColorPalette->GetValue( &red, &green, &blue, i  );	// Get

		*( pMemory + ( i * 3 ))			= blue;						// Set
		*( pMemory + ( i * 3 ) + 1 )	= green;
		*( pMemory + ( i * 3 ) + 2 )	= red;
	}
}

/********************************************************************************************************
	TXLibUtility::CTga::ReadColorIndexImage()

	Input memory data to ColorIndexImage
*********************************************************************************************************/
bool
CTga::ReadColorIndexImage(
			TXLib::CColorIndexImage*	pColorIndexImage,			// Color index image
	const	u8*							pMemory,					// Input data pointer
			u16							width,						// Width
			u16							height,						// Height
			u8							dataLineType					// Image data line format
) 
{
	assert(( pColorIndexImage != NULL ) && ( pMemory != NULL ));

	pColorIndexImage->Create( width, height);

	for( u16 y=0; y<height; y++)
	{
		u16 yUp = y;
		if( ( dataLineType == kDataLineType_leftDown ) || (dataLineType == kDataLineType_rightDown) )
		{
			yUp = height - 1 - y;	// // Reflect the Y coordinate
		}

		for( u16 x=0; x<width; x++ )
		{
			u16 xRight = x;
			if( ( dataLineType == kDataLineType_rightDown ) || (dataLineType == kDataLineType_rightUp) )
			{
				xRight = width - 1 - x;	// // Reflect the X coordinate
			}

			u8 index = *( pMemory + ( yUp * width ) + xRight     );	// Get
			pColorIndexImage->SetValue( index, x, y  );			// Set
		}
	}

	return true;
}

/********************************************************************************************************
	TXLibUtility::CTga::WriteColorIndexImage()

	Output ColorIndexImage to memory
*********************************************************************************************************/
bool
CTga::WriteColorIndexImage(
			u8*							pMemory,						// Output data pointer
	const	TXLib::CColorIndexImage*	pColorIndexImage,				// Color index image
			u16							bitCount,						// Bit depth
			u8							dataLineType					// Image data line format
) const
{
	assert( ( pMemory != NULL ) &&  ( pColorIndexImage != NULL ));
	assert(	pColorIndexImage->IsAvailable());

	if( bitCount != 8 ) return false;

	u16 width	= pColorIndexImage->GetWidth();
	u16 height	= pColorIndexImage->GetHeight();

	for( u16 y=0; y<height; y++)
	{
		u16 yUp = y;
		if( ( dataLineType == kDataLineType_leftDown ) || (dataLineType == kDataLineType_rightDown) )
		{
			yUp = height - 1 - y;	// // Invert Y coordinate
		}

		for( u16 x=0; x<width; x++ )
		{
			u16 xRight = x;
			if( ( dataLineType == kDataLineType_rightDown ) || (dataLineType == kDataLineType_rightUp) )
			{
				xRight = width - 1 - x;	// // Invert X coordinate
			}

			u8 index;

			pColorIndexImage->GetValue( &index, x, y  );	 // Get
			*( pMemory + ( yUp * width) + xRight	  ) = index;	// Set
		}
	}

	return true;
}

/********************************************************************************************************
	TXLibUtility::CTga::ReadAlphaPalette()

	Put the memory data into AlphaImage. Only when the bitcount is 32bit palette.
*********************************************************************************************************/
void
CTga::ReadAlphaPalette(
			TXLib::CAlphaImage*			pAlphaImage,					// Alpha image pointer
	const	u8*							pMemory,						// Input data pointer
			TXLib::CColorIndexImage*	pColorIndexImage				// Color index image
)
{
	assert(( pAlphaImage != NULL ) && ( pMemory != NULL ) && ( pColorIndexImage != NULL ));

	u16 width = pColorIndexImage->GetWidth();
	u16 height = pColorIndexImage->GetHeight();

	pAlphaImage->Create( width, height, 0xff);

	u8 index = 0;
	u8 alpha = 0;

	for( u16 y=0; y<height; y++)
	{
		for( u16 x=0; x<width; x++ )
		{
			pColorIndexImage->GetValue( &index, x, y );
			alpha = *( pMemory + index * 4 + 3 );
			pAlphaImage->SetValue( alpha, x, y ); // Alpha 
		}
	}
}

/*********************************************************************************************************
		public
*********************************************************************************************************/
/********************************************************************************************************
	TXLibUtility::CTga::Read()

	Load TGA data from memory
*********************************************************************************************************/
bool
CTga::Read(
			TXLib::CTexture*		pTexture,						// Output CTexture pointer
	const	u8*						pMemory							// Input data pointer
)
{
	assert((pTexture != NULL ) && ( pMemory != NULL ));

	tFileHeader header;

	// Read header
	if( !ReadTGAHEADER( &header, pMemory  ) ) return false;

	u16 width			= (u16)header.width;
	u16 height			= (u16)header.height;
	u32 nColorUsed		= 0;
	u8	dataLineType	= header.imageBit & 0x30;	// Get data line

	const u8*	pImageMemory;			// Image memory pointer
	u8*	pUnpackImageMemory	= NULL;		// Unpack data

	// Generate CBmp elements based on bit depth
	if( ( header.imageType == kImageType_palette ) || ( header.imageType == kImageType_paletteRle ) )
	{ // Color index 
		if( header.bitCount != 8 ) return false;

		pTexture->SetType( TXLib::CTexture::kType_palette256 );				// Set the CTexture format

		if( (u32)header.paletteColorNumber )
		{
			nColorUsed = (u32)header.paletteColorNumber;
		}
		else
		{
			nColorUsed = 256;
		}
		pImageMemory = (pMemory + kDefine_headerByteSize + header.idFieldByteSize ) + nColorUsed * (header.paletteBitCount / 8);

		// If there is RLE compression, allocate memory again and unpack original data 
		if( header.imageType == kImageType_paletteRle )
		{
			u32 unpackMemorySize = width * height * (header.bitCount >> 3 );	// Memory allocation for unpack data
			pUnpackImageMemory = new u8[ unpackMemorySize ];

			UnpackRLEData( pUnpackImageMemory, pImageMemory, header.bitCount, width, height );			// Unpack
			pImageMemory = pUnpackImageMemory;	// Change the image memory pointer to the unpacked data
		}

		ReadColorIndexImage	( &pTexture->GetColorIndexImage(), pImageMemory, width, height, dataLineType ); // Create color index image
		ReadColorPalette( &pTexture->GetColorPalette(), pMemory + kDefine_headerByteSize, nColorUsed, header.paletteBitCount );	// Create color palette

		if( header.paletteBitCount == 32)
		{
			ReadAlphaPalette( &pTexture->GetAlphaImage(), pMemory + kDefine_headerByteSize, &pTexture->GetColorIndexImage() );	// Create alpha image
		}
	}
	else if( ( header.imageType == kImageType_direct ) || ( header.imageType == kImageType_gray ) ||
		( header.imageType == kImageType_directRle ) || ( header.imageType == kImageType_grayRle ))
	{ // direct5551, direct8888, GRAY
		if( ( header.bitCount != 8 ) && ( header.bitCount != 16 ) &&
				( header.bitCount != 24 ) &&  ( header.bitCount != 32 ) ) return false;
	
		pTexture->SetType( TXLib::CTexture::kType_direct);						// Set CTexture format
		pImageMemory = pMemory + kDefine_headerByteSize + header.idFieldByteSize;

		// If there is RLE compression, allocate memory again and unpack original data 
		if( ( header.imageType == kImageType_directRle ) || ( header.imageType == kImageType_grayRle ) )
		{
			u32 unpackMemorySize = width * height * (header.bitCount >> 3 );	// Memory allocation for unpack data
			pUnpackImageMemory = new u8[ unpackMemorySize ];

			UnpackRLEData( pUnpackImageMemory, pImageMemory, header.bitCount, width, height );			// Unpack
			pImageMemory = pUnpackImageMemory;	// Change the image memory pointer to the unpacked data
		}

		ReadDirectColorImageAlpha(&pTexture->GetDirectColorImage(), &pTexture->GetAlphaImage(), pImageMemory,
									header.bitCount, width, height, dataLineType ); // Create direct color image
	}
	else
	{
		return false;
	}

	if( pUnpackImageMemory != NULL )
	{
		delete[] pUnpackImageMemory;		// Release the unpack data memory 
	}

	return true;
}

/********************************************************************************************************
	TXLibUtility::CTga::CTga::Write()

	Write Tga data to memory

	Write the correct bytes size if pByteSize exists
	The image data line creates Pixel: left to right, Row: top to bottom 
	ID field is "0"
	The image start position is at the origin (0,0)
*********************************************************************************************************/
bool
CTga::Write(
			u8*						pMemory,					// Output data pointer
			u32*					pByteSize,					// Number of bytes output
			TXLib::CTexture*		pTexture,					// CTexture pointer
			u8						dataLineType				// Image data line format
) const
{
	assert(( pMemory != NULL ) && (pTexture != NULL ));

	u32				nColorUsed;								// Number of colors used in the color palette
	u16				width		= pTexture->GetWidth();		// Width
	u16				height		= pTexture->GetHeight();	// Height

	TXLib::CTexture::tType	convertType		= GetConvertType( pTexture );			// CTexture format

	TXLib::CTexture texture;
	texture = *pTexture;		// Copy so that the original data is not converted

	u32	memoryDataOffset = 0;	// Memory offset

	if( convertType != pTexture->GetType())
	{
		if( !texture.Convert( convertType )) return false;								// Convert
	}

	u8 bitCount=0;
	if( convertType == TXLib::CTexture::kType_palette256 )
	{
		bitCount		= 8;
		nColorUsed		= pTexture->GetColorPalette().GetNumber();

		WriteTGAHEADER( pMemory, kImageType_palette, bitCount, width, height, nColorUsed, dataLineType );	// Header write
		memoryDataOffset += kDefine_headerByteSize;

		WriteColorPalette( pMemory + memoryDataOffset, &pTexture->GetColorPalette() );			// Palette write
		memoryDataOffset += nColorUsed * 3;

		if(!WriteColorIndexImage( pMemory + memoryDataOffset, &pTexture->GetColorIndexImage(), 8, dataLineType)) return false;	// Color index write
		memoryDataOffset += width * height * ( bitCount / 8 );
	}
	else if( convertType == TXLib::CTexture::kType_direct )
	{
		bitCount		= 32;

		WriteTGAHEADER( pMemory, kImageType_direct, bitCount, width, height, 0, dataLineType );		// Header write
		memoryDataOffset += kDefine_headerByteSize;

		if(!WriteDirectColorImageAlpha( pMemory + memoryDataOffset, &texture.GetDirectColorImage(),
				&texture.GetAlphaImage(), bitCount, dataLineType )) return false;	// Image write
		memoryDataOffset += width * height * ( bitCount / 8 );
	}
	else
	{
		return false;
	}

	WriteFooter( pMemory + memoryDataOffset, 0, 0 );	// Footer write
	memoryDataOffset += kDefine_footerByteSize;

	if( pByteSize != NULL )
	{
		*pByteSize = memoryDataOffset;
	}

	return true;
}

/********************************************************************************************************
	TXLibUtility::CTga::GetConvertType()

	Get the conversion CTexture format when TGA file is saved.

		CTexture format				TGA data
		TXLib::CTexture::kType_palette4		-> palette256
		TXLib::CTexture::kType_palette16	-> palette256
		TXLib::CTexture::kType_palette256	-> palette256
		TXLib::CTexture::kType_tex4x4		-> direct
		TXLib::CTexture::kType_direct		-> direct
		TXLib::CTexture::kType_a3i5			-> direct
		TXLib::CTexture::kType_a5i3			-> direct
*********************************************************************************************************/
TXLib::CTexture::tType
CTga::GetConvertType(
			TXLib::CTexture*		pTexture						// CTexture pointer
) const
{
	assert( pTexture != NULL);

	TXLib::CTexture::tType		convertType;		// Conversion CTexture format

	switch( pTexture->GetType() )
	{
		case TXLib::CTexture::kType_palette4:
		case TXLib::CTexture::kType_palette16:
		case TXLib::CTexture::kType_palette256:
			convertType	= TXLib::CTexture::kType_palette256;	break;
		case TXLib::CTexture::kType_tex4x4:
		case TXLib::CTexture::kType_direct:
		case TXLib::CTexture::kType_a5i3:
		case TXLib::CTexture::kType_a3i5:
			convertType	= TXLib::CTexture::kType_direct;	break;
		default:
			assert(false);
	}

	return convertType;
}

/********************************************************************************************************
	TXLibUtility::CTga::GetDataByteSizeMax()

	Get the largest file size necessary to save the TGA file.
*********************************************************************************************************/
bool
CTga::GetDataByteSizeMax(
		u32*					pMemorySize,					// Memory size
		TXLib::CTexture*		pTexture						// CTexture pointer
) const
{
	assert(( pMemorySize != NULL ) && (pTexture != NULL ));

	u16 width	= pTexture->GetWidth();
	u16 height	= pTexture->GetHeight();

	TXLib::CTexture::tType		convertType = GetConvertType( pTexture );		// Get CTexture format at time of save

	u32 memorySize = 0;

	if( convertType == TXLib::CTexture::kType_palette256 )
	{
		*pMemorySize = kDefine_headerByteSize + kDefine_footerByteSize + width * height + 256 * 3;
	}
	else if( convertType == TXLib::CTexture::kType_direct )
	{
		*pMemorySize = kDefine_headerByteSize + kDefine_footerByteSize + width * height * 4;
	}
	else
	{
		return false;
	}

	return true;
}

} // namespace
