/****************************************************************************
  Project: NITRO-SampleTools - TXLib
  File:    convertTextureType.cpp

  Copyright 2004,2005 Nintendo Co.,Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  - The input image file, texture format to be converted, NTF file name, and option output file name

  - Input one fo the following as the texture format.
	Input			Format
	0			No format conversion
	4			palette4
	8			a5i3
	16			palette16
	32			a3i5
	256			palette256
	24			direct
	44			tex4x4	

  - When the texture format is tex4x4, input a conversion coefficient.

   (Execution example)
   Converting BMP to tex4x4 and NTF and outputting.

  convertTextureType ../common/inputFiles/test_128x128_direct.tga 44 outputFiles/after.c outputFiles/after.bmp

  $Log: convertTextureType.cpp,v $
  Revision 1.8  2005/06/20 06:54:11  mominoki
  Changed copyright. Made revision adjustments. 

  Revision 1.1  2004/04/15		mominoki
  Corrected comment

  Revision 1.0  2004/03/03		mominoki
  Initial registration

  $NoKeywords: $
 ***************************************************************************/
#include "stdafx.h"
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <TXLib.h>
#include "../TXLibUtility/include/TXLibUtility.h"

#define LINE_CHAR	256

int _tmain(int argc, _TCHAR* argv[])
{
	TXLib::CTexture		texture;
	TXLib::CDefaultColorProcessor::tConvertFactor	factor;

	std::memset( (void*)&factor, 0, sizeof(TXLib::CDefaultColorProcessor::tConvertFactor));

	// Convert input image file To CTexture.
	if( !TXLibUtility::Load(
				&texture,											// CTexture
				argv[1]												// input image file path name
	) )
	{
		printf("Failed to load image file.\n"); exit(1);
	}

	TXLib::CTexture::tType	textureFormat;
	char					typeStr[LINE_CHAR];
	int						mode;

	// Get the CTexture format to be converted
	switch( atoi(argv[2]) )
	{
		case 0:
					textureFormat = texture.GetType();
					strcpy( typeStr, "format");
					break;
		case 4:
		case 16:
		case 256:
					// Use the Color Value in color0 of the palette
					factor.transparencyFlg = false;

					if( atoi(argv[2]) == 4 )
					{
						textureFormat = TXLib::CTexture::kType_palette4;
						strcpy( typeStr, "palette4");
					}
					else if( atoi(argv[2]) == 16 )
					{
						textureFormat = TXLib::CTexture::kType_palette16;
						strcpy( typeStr, "palette16");
					}
					else if( atoi(argv[2]) == 256 )
					{
						textureFormat = TXLib::CTexture::kType_palette256;
						strcpy( typeStr, "palette256");
					}
					break;
		case 8:
					textureFormat = TXLib::CTexture::kType_a5i3;
					strcpy( typeStr, "a5i3");
					break;
		case 32:
					textureFormat = TXLib::CTexture::kType_a3i5;
					strcpy( typeStr, "a3i5");
					break;
		case 24:
					textureFormat = TXLib::CTexture::kType_direct;
					strcpy( typeStr, "direct");
					break;
		case 44:
				{
					int		red, green, blue;

					// Create conversion coefficient
					printf(	"Select an algorithm. ( 0: Fast  Without linear interpolation,\n"
							"                       1: Fast  With linear interpolation,\n"
							"                       2: Round robin  Without linear interpolation,\n"
							"                       3: Round robin  With linear interpolation )\n");

					scanf( "%d", &mode );
					factor.algorithm = (TXLib::CDefaultColorProcessor::tAlgorithm)mode;

					printf("Select a mode. (0: Transparent Color Mode, 1: 4-Color Mode)\n");
					scanf( "%d", &mode );
					factor.transparencyFlg = (mode == 0) ? true : false;

					if( factor.transparencyFlg )
					{
						printf("Use transparent color? Use alpha? (0: Transparent color, 1: Alpha)\n");
						scanf("%d", &mode );
						factor.alphaColorCutFlg = (mode == 0) ? false : true;

						if( !factor.alphaColorCutFlg )
						{
							printf("Specify the transparent color in RGB order. (Red: 255 0 0)\n");
							scanf("%d %d %d", &red, &green, &blue );
							factor.transparentColor.red		= (u8)red;
							factor.transparentColor.green	= (u8)green;
							factor.transparentColor.blue	= (u8)blue;
						}
					}

					printf("Share palettes and compress the number of palettes? (0: No, 1: Yes)\n");
					scanf( "%d", &mode );
					factor.compressFlg = (mode == 0) ? false: true;

					if( factor.compressFlg )
					{
						printf("Specify the color region for palette sharing in RGB order. (0 - 31)\n"
								"If all values are set to 0, only palettes having the exact same colors will be shared.\n");
						scanf( "%d %d %d", &red, &green, &blue );
						factor.commonColorRange.red		= red;
						factor.commonColorRange.green	= green;
						factor.commonColorRange.blue	= blue;
					}

					textureFormat = TXLib::CTexture::kType_tex4x4;
					strcpy( typeStr, "tex4x4");
				}
				break;
		default:
				printf("Conversion has failed. Verify the texture format to be converted.\n");
				exit(1);
	}

	// Set conversion coefficient
	texture.GetColorProcessor()->SetConvertFactor( &factor );

	// Convert
	if( !texture.Convert( textureFormat ))
	{
		printf("Failed to convert the texture format.\n"); exit(1);
	}

	// Output NTF
	if( !TXLibUtility::SaveNtf(
				argv[3],							// NTF File name
				&texture,							// CTexture
				false,								// Texture format
				argv[1],							// Input image file name
				false,								// Doesn't output file for TextureViewer
				true								// Adjust NTF size
	) )
	{
		printf("Failed to output NTF file.\n"); exit(1);
	}

	// Output image file
	if( !TXLibUtility::Save(
				argv[4],
				&texture
	) )
	{
		printf("Failed to output image file.\n"); exit(1);
	}

	// Display
	printf("input image file name: %s\n",argv[1] );
	printf("texture format:         %s\n",typeStr );
	printf("output NTF file name:  %s\n",argv[3] );
	printf("output image file name: %s\n\n",argv[4] );
	printf("Succeeded.\n\n");

	return 0;
}

