/****************************************************************************
  Project: NITRO-SampleTools - TXLib
  File:    inputData.cpp

	Copyright 2004,2005 Nintendo Co.,Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

	This is for acquiring input data from the options that have been input.

  $Log: inputData.cpp,v $
  Revision 1.9  2005/06/20 06:54:11  mominoki
  Changed copyright. Made revision adjustments. 

  Revision 1.4	2004/10/28		mominoki
  When converting from direct to a3i5/a5i3, we added the ability to select as an option -ta, which was previously always specified.  

  Revision 1.3	2004/10/27		mominoki
  Added pBgAlignmentPaletteNumberFlg argument to GetInputBgData(). 
  During BG output, added the ablility to select whether to output palettes with 16 colors when 4 bit, and 256 colors when 8 bit.  

  Revision 1.2	2004/06/25		mominoki
  Fixed the bug that caused incorrect color subtraction when -ta was not specified 
  during a conversion from direct to a3i5, a5i3. 
  Changed the program so that -ta is always specified inside the program. 

  Revision 1.1	2004/04/15		mominoki
  Revised comments.

  Revision 1.0	2004/03/03		mominoki
  Registration of Initial release. 

  $NoKeywords: $
 ***************************************************************************/
#include "stdafx.h"
#include "cassert"

#include "../TXLibUtility/include/TXLibUtility.h"
#include "../TXLibUtilityEx/include/TXLibUtilityEx_Bg.h"
#include "MyCUtility.h"
#include "inputData.h"

/****************************************************************************
	GetTextureType()

	Get the texture format from the texture format name.
	Return value: texture format, if error: kType_not
 ***************************************************************************/
TXLib::CTexture::tType 
GetTextureType(
	const	char*				pTextureTypeName
)
{
	TXLib::CTexture::tType type;

	if( strcmp(			pTextureTypeName, "palette4"	) == 0)	type = TXLib::CTexture::kType_palette4;
	else if( strcmp(	pTextureTypeName, "a5i3"		) == 0)	type = TXLib::CTexture::kType_a5i3;
	else if( strcmp(	pTextureTypeName, "palette16"	) == 0)	type = TXLib::CTexture::kType_palette16;
	else if( strcmp(	pTextureTypeName, "a3i5"		) == 0)	type = TXLib::CTexture::kType_a3i5;
	else if( strcmp(	pTextureTypeName, "palette256"	) == 0)	type = TXLib::CTexture::kType_palette256;
	else if( strcmp(	pTextureTypeName, "direct"		) == 0)	type = TXLib::CTexture::kType_direct;
	else if( strcmp(	pTextureTypeName, "tex4x4"		) == 0)	type = TXLib::CTexture::kType_tex4x4;
	else														type = TXLib::CTexture::kType_not;

	return type;
}

/****************************************************************************
	GetInputData()

	Get data input from the console.
***************************************************************************/
tErrorStatus
GetInputData( 
			char*							pInputPathName,				// Input image file path name
			char*							pOutputNtfPathName,			// Output NTF path name
			bool*							pNTFBinaryFlg,				// Output NTF binary format flag
			bool*							pNotAlignmentFlg,			// NTF size alignment flag (true: no alignment)
			char*							pOutputOptionPathName,		// Output option file path name
			char*							pPalettePathName,			// File path name used for the palette
			TXLib::CTexture::tType*			pConvertTextureType,		// Texture conversion format
			TXLib::CDefaultColorProcessor
				::tConvertFactor*			pFactor,					// Conversion factor
			_TCHAR*							argv[],						// Input data
	const	int*							pOption						// option location
)
{
	assert( pInputPathName &&  pOutputNtfPathName && pNTFBinaryFlg && pOutputOptionPathName &&
		pConvertTextureType && pFactor && argv && pOption );

	char*	pChar			= NULL;
	int		optionNumber	= 0;

	// Input image file path name
	if( ( optionNumber = pOption[ kOption_InputPathName ] ) == 0 )	return kErrorStatus_NotInputPathName;
	strcpy( pInputPathName, argv[ optionNumber ] );	// Get path name
	
	// Extension error determination
	if( ( MyCUtility::IsCorrectExtName( pInputPathName, ".bmp" ) !=  0) &&
			( MyCUtility::IsCorrectExtName( pInputPathName, ".tga" ) !=  0) )
	{
		return kErrorStatus_DifferentInputExtName;
	}

	// Output NTF path name
	if( pOption[ kOption_NotNtf ] > 0 )
	{
		pOutputNtfPathName = NULL;	// No path name
	}
	else
	{
		// Output if NTF output is valid.
		if( ( optionNumber = pOption[ kOption_NtfPathName ] ) == 0 )
		{
			strcpy( pOutputNtfPathName, pInputPathName );				// Copy input image file path name.
		}
		else
		{
			strcpy( pOutputNtfPathName, argv[ optionNumber + 1 ] );	// Get path name
		}
		// TXLibUtility::SaveNtf() automatically appends the extension. Binary (.bin) or text (.c).
		// Change the path name extension for output content display.
		if( pOption[ kOption_NtfBinary ] > 0 )
		{
			TXLibUtility::ChangeExtName( pOutputNtfPathName, "." );
			*pNTFBinaryFlg = true;
		}
		else
		{
			TXLibUtility::ChangeExtName( pOutputNtfPathName, ".c" );
			*pNTFBinaryFlg = false;
		}

		// Determine whether the output NTF path name is valid for a C program.
		if( !MyCUtility::IsCorrectPathName( pOutputNtfPathName ) ) 	return kErrorStatus_FalseNtfPathName;
	}

	// Align NTF image size
	if( pOption[ kOption_NotAlignmentNtfSize ] > 0 )
	{
		*pNotAlignmentFlg = false;
	}
	else
	{
		*pNotAlignmentFlg = true;
	}

	// Output option file path name
	if( ( optionNumber = pOption[ kOption_OptionPathName ] ) != 0 )
	{
		// Specified
		strcpy( pOutputOptionPathName, argv[ optionNumber + 1 ] );	// Path name acquisition

		// Extension error determination
		if( ( MyCUtility::IsCorrectExtName( pOutputOptionPathName, ".bmp" ) !=	0) &&
				( MyCUtility::IsCorrectExtName( pOutputOptionPathName, ".tga" ) !=	0) )
		{
			return kErrorStatus_DifferentOptionExtName;
		}
	}

	// Palette file path name
	if( ( optionNumber = pOption[ kOption_PalettePathName ] ) != 0 )
	{
		// Specified
		strcpy( pPalettePathName, argv[ optionNumber + 1 ] );	// Path name acquisition

		// Extension error determination
		if( ( MyCUtility::IsCorrectExtName( pPalettePathName, ".bmp" ) !=  0) &&
				( MyCUtility::IsCorrectExtName( pPalettePathName, ".tga" ) !=  0) )
		{
			return kErrorStatus_DifferentPaletteExtName;
		}
	}


	// Texture conversion format
	if( ( optionNumber = pOption[ kOption_TextureFormat ] ) != 0 )
	{
		*pConvertTextureType = GetTextureType( argv[ optionNumber + 1 ] );	// Get conversion format
		if( *pConvertTextureType == TXLib::CTexture::kType_not ) return kErrorStatus_DifferentTextureType;
	}
	else
	{
		*pConvertTextureType = TXLib::CTexture::kType_not;	// Not specified
	}

	// Four color mode or transparent color mode
	if( pOption[ kOption_TransparencyAlpha ] > pOption[ kOption_TransparencyColor ] )
	{	// Transparent color mode,   Alpha extract
		pFactor->transparencyFlg = true;
		pFactor->alphaColorCutFlg = true;
	}
	else if( ( optionNumber =  pOption[ kOption_TransparencyColor ]) > 0 )
	{	// Transparent color mode,   Transparent color extract
		pFactor->transparencyFlg = true;

		// Read transparent color
		if( !MyCUtility::GetOptionColor( 
				&pFactor->transparentColor.red,
				&pFactor->transparentColor.green,
				&pFactor->transparentColor.blue,
				argv[ optionNumber ],
				",") ) return kErrorStatus_TransparencyColor;
	}
	else
	{	// Four color mode
		pFactor->transparencyFlg = false;
	}

	// Coefficient for tex4x4 conversion 
	if( *pConvertTextureType == TXLib::CTexture::kType_tex4x4 )
	{
		// Algorithm (Select the last algorithm that was specified)
		tOption		algorithmOption = kOption_Algorithm2;
		pFactor->algorithm = TXLib::CDefaultColorProcessor::kAlgorithm_roundRobinNotLinear;

		if( pOption[ kOption_Algorithm3 ] > pOption[ algorithmOption ] )
		{
			algorithmOption = kOption_Algorithm3;		// Default
			pFactor->algorithm = TXLib::CDefaultColorProcessor::kAlgorithm_roundRobinLinear;
		}
		if( pOption[ kOption_Algorithm0 ] > pOption[ algorithmOption ] )
		{
			algorithmOption = kOption_Algorithm0;
			pFactor->algorithm = TXLib::CDefaultColorProcessor::kAlgorithm_fastNotLinear;
		}
		if( pOption[ kOption_Algorithm1 ] > pOption[ algorithmOption ] )
		{
			algorithmOption = kOption_Algorithm1;
			pFactor->algorithm = TXLib::CDefaultColorProcessor::kAlgorithm_fastLinear;
		}

		// Palette number compression 
		if( pOption[ kOption_NotCompressPalette ] == 0 )
		{	// With compression 
			pFactor->compressFlg = true;

			if( ( optionNumber = pOption[ kOption_CompressPalette ]) > 0 )
			{
				// Read common color range
				if( !MyCUtility::GetOptionColor( 
						&pFactor->commonColorRange.red,
						&pFactor->commonColorRange.green,
						&pFactor->commonColorRange.blue,
						argv[ optionNumber ],
						",") ) return kErrorStatus_CompressPalette;

				if( ( pFactor->commonColorRange.red > 31 ) || ( pFactor->commonColorRange.green > 31 ) 
					|| ( pFactor->commonColorRange.blue > 31 ) ) return kErrorStatus_CompressPalette;
			}
		}
		else
		{	// No compression
			pFactor->compressFlg = false;
		}
	}

	// Display input data
	if( pOption[ kOption_View ] > 0 )
	{
		printf( "input image file name:               %s\n", pInputPathName );

		// NTF
		if( pOutputNtfPathName != NULL )
		{
			if( *pNTFBinaryFlg )
			{
				printf( "output NTF file name(binary format): %s(.ntft)(.ntfp)(.ntfi)\n", pOutputNtfPathName );
			}
			else
			{
				printf( "output NTF file name(text format):   %s\n", pOutputNtfPathName );
			}
		}

		if( pOutputOptionPathName[0] != NULL )
		{
			printf( "output image file name:              %s\n", pOutputOptionPathName );
		}

		if( pPalettePathName[0] != NULL )
		{
			printf( "input palette file name:             %s\n", pPalettePathName );
		}

		optionNumber = pOption[ kOption_TextureFormat ];
		if( optionNumber>0 )
		{
			printf( "convert texture format:              %s\n", argv[ optionNumber + 1 ] );
		}

		// Transparent color 
		if( *pConvertTextureType == TXLib::CTexture::kType_tex4x4 )
		{
			if( !pFactor->transparencyFlg )
			{
				printf( "mode:                                4-Color Mode\n");
			}
			else
			{
				if( pFactor->alphaColorCutFlg )
				{
					printf( "mode:                                Transparent Color Mode(Use Alpha) \n");
				}
				else
				{
					printf( "mode:                                Transparent Color Mode (Use Transparent Color) \n");
					printf( "Transparent Color:                   red: %3d, green: %3d, blue: %3d\n", pFactor->transparentColor.red,
							pFactor->transparentColor.green, pFactor->transparentColor.blue );
				}
			}
		}
		else if( (*pConvertTextureType == TXLib::CTexture::kType_palette4) || (*pConvertTextureType == TXLib::CTexture::kType_palette16) ||
			(*pConvertTextureType == TXLib::CTexture::kType_palette256))
		{
			if( !pFactor->transparencyFlg )
			{
				printf( "palette number 0:                    Use Color\n");
			}
			else
			{
				if( pFactor->alphaColorCutFlg )
				{
					printf( "palette number 0:                    Transparent Color (Use Alpha) \n");
				}
				else
				{
					printf( "palette number 0:                    Transparent Color (Use Transparent Color) \n");
					printf( "Transparent Color:                   red: %3d, green: %3d, blue: %3d\n", pFactor->transparentColor.red,
									pFactor->transparentColor.green, pFactor->transparentColor.blue );
				}
			}
		}

		if( *pConvertTextureType == TXLib::CTexture::kType_tex4x4 )
		{
			printf( "algorithm:                           %d\n", (int)pFactor->algorithm);

			if( pFactor->compressFlg )
			{
				printf( "compress the number of palettes:     Yes\n");
				printf( "shared color range to compress\n"
					"the number of palettes:              red: %3d, green: %3d, blue: %3d\n", pFactor->commonColorRange.red,
					pFactor->commonColorRange.green, pFactor->commonColorRange.blue );
			}
			else
			{
				printf( "compress the number of palettes:     No\n");
			}
		}
	}

	return kErrorStatus_Success;
}

#ifdef USE_BG
/****************************************************************************
	GetBgType()

	Get CBg format from CBg format name.
	Return value: CBg format, if error: kType_not
 ***************************************************************************/
TXLibUtilityEx::CBg::tType 
GetBgType(
	const	char*				pBgTypeName
)
{
	TXLibUtilityEx::CBg::tType	type;

	if( strcmp(			pBgTypeName, "palette16"	) == 0)	type = TXLibUtilityEx::CBg::kType_text_palette16;
	else if( strcmp(	pBgTypeName, "palette256"	) == 0)	type = TXLibUtilityEx::CBg::kType_text_palette256;
	else if( strcmp(	pBgTypeName, "affine"	) == 0)		type = TXLibUtilityEx::CBg::kType_affine;
	else if( strcmp(	pBgTypeName, "palette256x16") == 0)	type = TXLibUtilityEx::CBg::kType_exAffine_palette256x16;
	else if( strcmp(	pBgTypeName, "palette256Bmp") == 0)	type = TXLibUtilityEx::CBg::kType_exAffine_palette256Bmp;
	else if( strcmp(	pBgTypeName, "direct"	) == 0)		type = TXLibUtilityEx::CBg::kType_exAffine_direct;
	else if( strcmp(	pBgTypeName, "big256Bmp"	) == 0)		type = TXLibUtilityEx::CBg::kType_big256Bmp;
	else													type = TXLibUtilityEx::CBg::kType_not;

	return type;
}

/****************************************************************************
	GetInputBgData()

	Get Bg data that was input from the console.
	Call this function immediately after GetInputData().
***************************************************************************/
tErrorStatus
GetInputBgData( 
			char*							pOutputBgPathName,			// Output Bg path name
			bool*							pBgBinaryFlg,				// Output Bg binary format flag
			TXLibUtilityEx::CBg::tType*		pType,						// BG format
			u8*								pBgPaletteNumber,			// Palette number
			bool*							pBgAlignmentPaletteNumberFlg,	// Palette number alignment flag (On: true; Off: false)
			u8*								pOffsetImage,				// Color index offset value 
			bool*							pVFlipFlg,					// VFlip flag (On: true, Off: false)
			bool*							pHFlipFlg,					// HFlip flag (On: true, Off: false) 
			bool*							pCompressFlg,				// Character compression flag (On: true, Off: false)
	const	char*							pInputPathName,				// Input image file path name
			_TCHAR*							argv[],						// Input data
	const	int*							pOption						// Position of option
)
{
	assert( pOutputBgPathName &&  pBgBinaryFlg && pBgPaletteNumber && pVFlipFlg &&
		pHFlipFlg && pInputPathName && argv && pOption );


	char*	pChar			= NULL;
	int		optionNumber	= 0;

	// If there is no BG output, nothing
	if( pOption[ kOption_Bg ] == 0 )
	{
		*pOutputBgPathName = NULL;
		return kErrorStatus_Success;
	}

	// BG path name
	if( ( optionNumber = pOption[ kOption_BgPathName ] ) == 0 )
	{
		strcpy( pOutputBgPathName, pInputPathName );			// Copy input image file path name
	}
	else
	{
		strcpy( pOutputBgPathName, argv[ optionNumber + 1 ] );	// Get path name
	}
	// TXLibUtility::SaveBg() automatically appends extension. Binary format (.nbfs, .nbfc, nbfp) or text format (.c).
	// Change the path name extension for output content display.
	if( pOption[ kOption_BgBinary ] > 0 )
	{
		TXLibUtility::ChangeExtName( pOutputBgPathName, "." );
		*pBgBinaryFlg = true;
	}
	else
	{
		TXLibUtility::ChangeExtName( pOutputBgPathName, ".c" );
		*pBgBinaryFlg = false;
	}

	// Determine whether the output NTF path name is valid for a C program.
	if( !MyCUtility::IsCorrectPathName( pOutputBgPathName ) ) 	return kErrorStatus_FalseBgPathName;

	// BG format
	if( ( optionNumber = pOption[ kOption_BgFormat ] ) != 0 )
	{
		*pType = GetBgType( argv[ optionNumber + 1 ] );	// Get conversion format
		if( *pType == TXLibUtilityEx::CBg::kType_not ) return kErrorStatus_DifferentBgType;
	}
	else
	{  
		*pType = TXLibUtilityEx::CBg::kType_not;	// Not specified
	}


	// H-Flip,V-Flip
	*pHFlipFlg = ( pOption[ kOption_BgNotHFlip ] > 0 ) ? false: true;
	*pVFlipFlg = ( pOption[ kOption_BgNotVFlip ] > 0 ) ? false: true;

	// Character compression
	*pCompressFlg = ( pOption[ kOption_BgNotCompressChar ] > 0 ) ? false: true;

	// Palette number
	if( ( optionNumber = pOption[ kOption_BgPaletteNumber ] ) > 0 )
	{
		*pBgPaletteNumber = (u8)strtol( argv[ optionNumber + 1], &pChar, 0);	// Read the value
		if(( pChar == argv[ optionNumber + 1] )
			|| ( *pBgPaletteNumber > 15 )) return kErrorStatus_BgPaletteNumber;	// Not a number or more than 16
	}
	else
	{
		*pBgPaletteNumber = 0;
	}

	// Palette number alignment flag 
	*pBgAlignmentPaletteNumberFlg = ( pOption[ kOption_BgNotAlignmentPaletteNumber ] > 0 ) ? false: true;

	// Color index offset value
	if( ( optionNumber = pOption[ kOption_BgOffsetImage ] ) > 0 )
	{
		*pOffsetImage = (u8)strtol( argv[ optionNumber + 1], &pChar, 0);	// Read the value
		if( pChar == argv[ optionNumber + 1]  ) return kErrorStatus_BgOffsetImage;	// Not a number
	}
	else
	{
		*pOffsetImage = 0;
	}

	// Display input data 
	if( pOption[ kOption_View ] > 0 )
	{
		if( *pBgBinaryFlg )
		{
			printf( "output Bg File Name(Binary Format): %s(.nbfp)(.nbfc)(.nbfs)\n", pOutputBgPathName );
		}
		else
		{
			printf( "output Bg File Name(Text Format):   %s\n", pOutputBgPathName );
		}

		// Format
		if( ( optionNumber = pOption[ kOption_BgFormat ]) > 0 )
		{
			printf( "Bg Format:                          %s\n", argv[ optionNumber + 1 ] );
		}
	}

	return kErrorStatus_Success;
}
#endif // USE_BG


#include "ntexconv.h"


