/****************************************************************************
  Project: NITRO-SampleTools - TXLib
  File:    TXLib_DefaultColorProcessor.h

  Copyright 2004,2005 Nintendo Co.,Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  This class conducts color processing. It is called fromTXLib::CTexture.
  Call this class in the initial state of CTexture.
  Similar classes have been each prepared to call the other TXLib_ColorProcessor.

  $Log: TXLib_DefaultColorProcessor.h,v $
  Revision 1.7  2005/06/20 06:54:11  mominoki
  Updated copyright. Made revision adjustments. 

  Revision 1.1  2004/04/15		mominoki
  Revised the comments.

  Revision 1.0  2004/03/03		mominoki
  Initial registration.

  $NoKeywords: $
 ***************************************************************************/
#ifndef TXLIB_DEFAULTCOLORPROCESSOR_H_
#define TXLIB_DEFAULTCOLORPROCESSOR_H_

#include <TXLib_Typedef.h>
#include <TXLib/Convert/TXLib_ColorProcessor.h>
#include <TXLib/Base/TXLib_ColorPalette.h>
#include <TXLib/Base/TXLib_Texture.h>

namespace TXLib
{
#ifdef TLB_EXPORT_
	class __declspec(dllexport)CDefaultColorProcessor :public CColorProcessor
#else
	class __declspec(dllimport)CDefaultColorProcessor :public CColorProcessor
#endif
	{
	public:
		// Algorithm
		enum tAlgorithm
		{
			kAlgorithm_fastNotLinear,					// High speed   No linear interpolation
			kAlgorithm_fastLinear,						// High speed   Linear interpolation
			kAlgorithm_roundRobinNotLinear,				// Round-robin   No linear interpolation
			kAlgorithm_roundRobinLinear					// Round-robin   Linear interpolation
		};

		// Conversion Factor
		typedef struct tConvertFactor
		{
			bool					transparencyFlg;			// Transparent color mode validity flag ( false: color value   true: transparent color mode)
			tAlgorithm				algorithm;			// Algorithm
			tTLBRGB 				transparentColor;		// Transparent color
			bool	                        alphaColorCutFlg;             // Cut the transparent color with the alpha ( false: cut with transparent color   true: cut with alpha)
			tTLBRGB 				commonColorRange;	// Palette common range value
			bool					compressFlg;				// Palette compression flag ( false: no compression   true: compression )
			TXLib::CColorPalette*	pUsePalette;				// Palette used in conversion from direct to palette format
		}tConvertFactor;

		tConvertFactor				m_convertFactor;			// Conversion factor

	public:
		CDefaultColorProcessor();								// Constructor

		virtual		~CDefaultColorProcessor(){};				// Destructor

		tConvertFactor		GetConvertFactor()		const	{ return m_convertFactor; }		// Get the conversion factor

		//------------------------------------------------------------------------------------------
		// Using both palette format and tex4x4
		//------------------------------------------------------------------------------------------
		int CDefaultColorProcessor::FindNearPalette(
			const	CColorPalette*				pDstColorPalette,				// Palette to search
			const	CColorPalette*				pSrcColorPalette,				// Search target palette
					u32							nSrcColor,						// Number of colors to search
					u8							commonRed,						// Shared color range (red)
					u8							commonGreen,					// Shared color range (green)
					u8							commonBlue						// Shared color range (blue)
		); // Searches for a palette with the same layout from the palette


		bool CreateColorMemory(
					CColorPalette*				pColorPalette,					// Palette to be newly created
			const	CDirectColorImage*			pDirectColorImage,				// Direct color image
			const	CAlphaImage*				pAlphaImage,					// Alpha
					u8							borderValue						// Border value that splits the alpha value into 2 values (This value and above are nontransparent)
		); // Create the color memory from the direct color image


		//------------------------------------------------------------------------------------------
		// Used in palette format texture creation
		//------------------------------------------------------------------------------------------
		void CompressPalette_MedhianCut( 
					CColorPalette*				pNewColorPalette,				// Palette to be newly created
			const	CColorPalette* 				pOldColorPalette				// Current color palette
		); // Compress the number of colors with the median count method

		//------------------------------------------------------------------------------------------
		// Used in tex4x4 creation
		//------------------------------------------------------------------------------------------
		bool SetPalette_Tex4x4( 
					CColorPalette*				pNewColorPalette,				// Palette to be newly created
			const	CColorPalette* 				pOldColorPalette,				// Current color palette
					bool						transparencyFlg					// Transparent color mode flag
		); // Sets a pallete of up to 4 colors in a tex4x4 palette


		bool Get4ColorHiLow_Tex4x4( 
					CColorPalette*				pNewColorPalette,				// Palette to be newly created
			const	CColorPalette* 				pOldColorPalette,				// Current color palette
					bool						transparencyFlg,				// Transparent color mode flag
					bool						linearFlg						// Linear interpolation flag
		); // Creates a tex4x4 palette with the highest, lowest, and median brightness.


		bool Get4ColorRoundRobin_Tex4x4( 
					CColorPalette*				pNewColorPalette,				// Palette to be newly created
			const	CColorPalette* 				pOldColorPalette,				// Current color palette
					bool						transparencyFlg,				// Transparent color mode flag
					bool						linearFlg						// Linear interpolation flag
		); // Compresses the number of colors using the round-robin method, and creates a tex4x4 palette.


		void MakeBlockDirectColorImage_Tex4x4(
					CDirectColorImage*			pBlockDirectColorImage,			// 1 block image
			const	CDirectColorImage*			pSrcDirectColorImage,			// Original image
					u16							x,								// Starting location of original image block (X-coordinate)
					u16							y								// Starting location of original image block (Y-coordinate)
		); // Creates a 1 block (4x4 texel) direct color image from the direct color image.


		bool MakeBlockAlphaImageUseTransparentColor_Tex4x4(
					CAlphaImage*				pBlockAlphaImage,				// 1 block alpha
			const	CDirectColorImage*			pBlockDirectColorImage,			// 1 block image
					tTLBRGB						transparentColor				// Transparent color
		); // Creates a 1 block alpha image from the transparent color


		bool MakeBlockAlphaImageUseAlpha_Tex4x4(
					CAlphaImage*				pBlockAlphaImage,				// 1 block alpha
			const	CAlphaImage*				pSrcAlphaImage,					// Original alpha
					u16							x,								// Starting location of original alpha block (X-coordinate)
					u16							y,								// Starting location of original alpha block (Y-coordinate)
					u8							borderValue						// Border value that splits the alpha value into 2 values
		); // Creates a 1 block alpha image from the alpha image


		bool Get4Color_Tex4x4(
					CColorPalette*				pColorPalette,					// Color palette
					bool						transparencyFlg,				// Transparent color mode flag
			const	CDirectColorImage*			pBlockDirectColorImage,			// 1 block image
			const	CAlphaImage*				pBlockAlphaImage,				// 1 block alpha
					tAlgorithm					algorithm						// Algorithm
		); // Gets the tex4x4 palette from 1 block


		void ConvertLinearColorPalette_Tex4x4(
					CColorPalette*				pColorPalette,					// 2-color color palette
					bool						transparencyFlg					// Transparent color mode flag
		); // Creates a tex4x4 linearly interpolated palette from the 2-color

		bool CompressColorNumber_Tex4x4( 
					CColorPalette*				pColorPalette,					// Palette
					CByteData*					pColorPaletteAddress,			// Palette address
			const	CData*						pColorPaletteAttribute,			// Palette attributes
					u8							commonRed,						// Common region value   Red
					u8							commonGreen,					// Common region value   Green
					u8							commonBlue						// Common region value   Blue
		); // Compresses the number of palettes

		//------------------------------------------------------------------------------------------
		// virtual
		//------------------------------------------------------------------------------------------
		virtual	void SetConvertFactor( const void*	pFactor )
					{ m_convertFactor = *((tConvertFactor*)pFactor);};		// Sets the conversion factor


		virtual int FindNearColor(
			const	CColorPalette*				pPalette,					// Search target palette
					u8							red,
					u8							green,
					u8							blue,
					s16							avoidNumber					// Index that avoids transparent colors, etc.
		); 	// Finds nearest color


		virtual bool ConvertToPaletteType(
					CTexture*					pTexture					// Texture
		); 	// Converts to the palette format texture from direct


		virtual bool ConvertToTex4x4(
					CTexture*					pTexture					// Texture
		); 	// Converts to tex4x4 from direct
	};
};

#endif